from __future__ import annotations

import json
import logging
import os
import urllib
import uuid
import webbrowser
from threading import Semaphore, Thread, main_thread

import AppKit
import Foundation
import WebKit
from objc import nil, super
from PyObjCTools import AppHelper

from webview import FileDialog, _state, windows
from webview import settings as webview_settings
from webview.dom import _dnd_state
from webview.menu import Menu, MenuAction, MenuSeparator
from webview.models import Request, Response
from webview.screen import Screen
from webview.util import (
    DEFAULT_HTML,
    create_cookie,
    inject_pywebview,
    js_bridge_call,
    parse_file_type,
    stringify_headers,
)
from webview.window import FixPoint

# This lines allow to load non-HTTPS resources, like a local app as: http://127.0.0.1:5000
bundle = AppKit.NSBundle.mainBundle()
info = bundle.localizedInfoDictionary() or bundle.infoDictionary()
info['NSAppTransportSecurity'] = {'NSAllowsArbitraryLoads': Foundation.YES}
info['NSRequiresAquaSystemAppearance'] = Foundation.NO  # Enable dark mode support for Mojave

# Fallbacks, in case these constants are not wrapped by PyObjC
try:
    NSFullSizeContentViewWindowMask = AppKit.NSFullSizeContentViewWindowMask
except AttributeError:
    NSFullSizeContentViewWindowMask = 1 << 15

try:
    NSWindowTitleHidden = AppKit.NSWindowTitleHidden
except AttributeError:
    NSWindowTitleHidden = 1

logger = logging.getLogger('pywebview')
logger.debug('Using Cocoa')

renderer = 'wkwebview'


class BrowserView:
    instances = {}
    app = AppKit.NSApplication.sharedApplication()
    app.setActivationPolicy_(0)
    current_menu = None

    cascade_loc = Foundation.NSMakePoint(100.0, 0.0)
    _shared_app_delegate = None

    class AppDelegate(AppKit.NSObject):
        def applicationShouldTerminate_(self, app):
            should_close = True
            for i in BrowserView.instances.values():
                should_close = should_close and BrowserView.should_close(i.pywebview_window)
            return Foundation.YES if should_close else Foundation.NO

        def applicationSupportsSecureRestorableState_(self, app):
            return Foundation.YES

    class WindowHost(AppKit.NSWindow):
        def canBecomeKeyWindow(self):
            return self.focus

    class WindowDelegate(AppKit.NSObject):
        def windowDidBecomeKey_(self, notification):
            i = BrowserView.get_instance('window', notification.object())

            if i and i.menu and BrowserView.current_menu != i.menu:
                BrowserView.current_menu = i.menu
                new_menu = i._recreate_menus(BrowserView.current_menu)
                BrowserView.app.setMainMenu_(new_menu)

            if not i:
                logger.error('No BrowserView instance found for windowDidBecomeKey_')

        def windowShouldClose_(self, window):
            i = BrowserView.get_instance('window', window)
            return BrowserView.should_close(i.pywebview_window)

        def windowWillClose_(self, notification):
            i = BrowserView.get_instance('window', notification.object())

            # Clear delegates before removing from instances dict so no
            # callbacks can arrive after get_instance would return None.
            i.webview.setNavigationDelegate_(None)
            i.webview.setUIDelegate_(None)

            del BrowserView.instances[i.uid]

            if i.pywebview_window in windows:
                windows.remove(i.pywebview_window)

            # this seems to be a bug in WkWebView, so we need to load blank html
            # see https://stackoverflow.com/questions/27410413/wkwebview-embed-video-keeps-playing-sound-after-release
            i.webview.loadHTMLString_baseURL_('', None)
            i.webview.removeFromSuperview()
            i.webview = None

            i.closed.set()
            if BrowserView.instances == {}:
                BrowserView.app.setDelegate_(None)
                BrowserView._shared_app_delegate = None
                BrowserView.app.stop_(self)
                BrowserView.app.abortModal()

        def windowDidResize_(self, notification):
            i = BrowserView.get_instance('window', notification.object())

            if i:
                size = i.window.frame().size
                i.pywebview_window.events.resized.set(size.width, size.height)

        def windowDidMiniaturize_(self, notification):
            i = BrowserView.get_instance('window', notification.object())

            if i:
                i.pywebview_window.events.minimized.set()

        def windowDidDeminiaturize_(self, notification):
            i = BrowserView.get_instance('window', notification.object())

            if i:
                i.pywebview_window.events.restored.set()

        def windowDidEnterFullScreen_(self, notification):
            i = BrowserView.get_instance('window', notification.object())
            if i:
                i.pywebview_window.events.maximized.set()

        def windowDidExitFullScreen_(self, notification):
            i = BrowserView.get_instance('window', notification.object())
            if i:
                i.pywebview_window.events.restored.set()

        def windowDidMove_(self, notification):
            i = BrowserView.get_instance('window', notification.object())
            if i:
                frame = i.window.frame()
                screen = i.window.screen().frame()
                flipped_y = screen.size.height - frame.size.height - frame.origin.y
                i.pywebview_window.events.moved.set(frame.origin.x, flipped_y)

    class JSBridge(AppKit.NSObject):
        def initWithObject_(self, window):
            super(BrowserView.JSBridge, self).init()
            self.window = window
            return self

        def userContentController_didReceiveScriptMessage_(self, controller, message):
            body = json.loads(message.body())
            if body['params'] is WebKit.WebUndefined.undefined():
                body['params'] = None
            js_bridge_call(self.window, body['funcName'], body['params'], body['id'])

    class DownloadDelegate(AppKit.NSObject):
        # Download delegate to handle links with download attribute set
        def download_decideDestinationUsingResponse_suggestedFilename_completionHandler_(
            self, download, decideDestinationUsingResponse, suggestedFilename, completionHandler
        ):
            save_dlg = AppKit.NSSavePanel.savePanel()
            directory = Foundation.NSSearchPathForDirectoriesInDomains(
                Foundation.NSDownloadsDirectory, Foundation.NSUserDomainMask, True
            )[0]
            save_dlg.setDirectoryURL_(Foundation.NSURL.fileURLWithPath_(directory))
            save_dlg.setNameFieldStringValue_(suggestedFilename)
            if save_dlg.runModal() == AppKit.NSFileHandlingPanelOKButton:
                filename = save_dlg.filename()
                url = Foundation.NSURL.fileURLWithPath_(filename)
                completionHandler(url)
            else:
                completionHandler(None)

    class BrowserDelegate(AppKit.NSObject):
        # Display a JavaScript alert panel containing the specified message
        def webView_runJavaScriptAlertPanelWithMessage_initiatedByFrame_completionHandler_(
            self, webview, message, frame, handler
        ):
            AppKit.NSRunningApplication.currentApplication().activateWithOptions_(
                AppKit.NSApplicationActivateIgnoringOtherApps
            )
            alert = AppKit.NSAlert.alloc().init()
            alert.setInformativeText_(str(message))
            alert.runModal()

            handler()

        def webView_didReceiveAuthenticationChallenge_completionHandler_(
            self, webview, challenge, handler
        ):
            # Prevent `ObjCPointerWarning: PyObjCPointer created: ... type ^{__SecTrust=}`

            # this allows any server cert
            if webview_settings['IGNORE_SSL_ERRORS'] or _state['ssl']:
                credential = AppKit.NSURLCredential.credentialForTrust_(
                    challenge.protectionSpace().serverTrust()
                )
                handler(AppKit.NSURLSessionAuthChallengeUseCredential, credential)
            else:
                handler(AppKit.NSURLSessionAuthChallengePerformDefaultHandling, nil)

        # Display a JavaScript confirm panel containing the specified message
        def webView_runJavaScriptConfirmPanelWithMessage_initiatedByFrame_completionHandler_(
            self, webview, message, frame, handler
        ):
            i = BrowserView.get_instance('webview', webview)
            ok = i.localization['global.ok']
            cancel = i.localization['global.cancel']

            result = BrowserView.display_confirmation_dialog(ok, cancel, message)
            handler(result)

        # Display a Javascript input panel
        def webView_runJavaScriptTextInputPanelWithPrompt_defaultText_initiatedByFrame_completionHandler_(
            self, webview, prompt, default_text, frame, handler
        ):
            i = BrowserView.get_instance('webview', webview)
            ok = i.localization['global.ok']
            cancel = i.localization['global.cancel']

            result = BrowserView.display_input_dialog(ok, cancel, prompt, default_text)
            handler(result)

        # Display an open panel for <input type="file"> element
        def webView_runOpenPanelWithParameters_initiatedByFrame_completionHandler_(
            self, webview, param, frame, handler
        ):
            i = list(BrowserView.instances.values())[0]
            file_filter = param._acceptedMIMETypes()

            files = i.create_file_dialog(
                FileDialog.OPEN,
                '',
                param.allowsMultipleSelection(),
                '',
                file_filter,
                main_thread=True,
            )

            if files:
                urls = [Foundation.NSURL.fileURLWithPath_(i) for i in files]
                handler(urls)
            else:
                handler(nil)

        # Open target="_blank" links in external browser
        def webView_createWebViewWithConfiguration_forNavigationAction_windowFeatures_(
            self, webview, config, action, features
        ):
            if action.navigationType() == getattr(WebKit, 'WKNavigationTypeLinkActivated', 0):
                if webview_settings['OPEN_EXTERNAL_LINKS_IN_BROWSER']:
                    webbrowser.open(action.request().URL().absoluteString(), 2, True)
                else:
                    webview.loadRequest_(action.request())
            return nil

        # WKNavigationDelegate method, invoked when a navigation decision needs to be made
        def webView_decidePolicyForNavigationAction_decisionHandler_(
            self, webview, action, handler
        ):
            # The event that might have triggered the navigation
            event = AppKit.NSApp.currentEvent()

            # Handle links with the download attribute set to recommend a file name
            if action.shouldPerformDownload() and webview_settings['ALLOW_DOWNLOADS']:
                handler(getattr(WebKit, 'WKNavigationActionPolicyDownload', 2))
                return

            """ Disable back navigation on pressing the Delete key: """
            # Check if the requested navigation action is Back/Forward
            if action.navigationType() == getattr(WebKit, 'WKNavigationTypeBackForward', 2):
                # Check if the event is a Delete key press (keyCode = 51)
                if event and event.type() == AppKit.NSKeyDown and event.keyCode() == 51:
                    # If so, ignore the request and return
                    handler(getattr(WebKit, 'WKNavigationActionPolicyCancel', 0))
                    return

            request = action.request()
            url = request.URL()

            original_headers = dict(request.allHTTPHeaderFields())
            i = BrowserView.get_instance('webview', webview)

            if (
                len(i.pywebview_window.events.request_sent) > 0
                and 'X-Handled' not in original_headers
                and str(url) != 'about:blank'
            ):
                request_ = Request(str(url), request.HTTPMethod(), original_headers)
                i.pywebview_window.events.request_sent.set(request_)
                request_.headers['X-Handled'] = 'true'

                new_request = Foundation.NSMutableURLRequest.requestWithURL_(url)
                new_request.setHTTPMethod_(request.HTTPMethod())
                new_request.setAllHTTPHeaderFields_(
                    AppKit.NSDictionary(stringify_headers(request_.headers))
                )
                new_request.setHTTPBody_(request.HTTPBody())
                new_request.setHTTPBodyStream_(request.HTTPBodyStream())
                new_request.setHTTPShouldHandleCookies_(request.HTTPShouldHandleCookies())
                new_request.setHTTPShouldUsePipelining_(request.HTTPShouldUsePipelining())
                new_request.setMainDocumentURL_(request.mainDocumentURL())
                new_request.setNetworkServiceType_(request.networkServiceType())
                webview.loadRequest_(new_request)
                handler(getattr(WebKit, 'WKNavigationActionPolicyCancel', 1))
            else:
                # Normal navigation, allow
                handler(getattr(WebKit, 'WKNavigationActionPolicyAllow', 1))

        def webView_navigationAction_didBecomeDownload_(self, webview, navigationAction, download):
            download.setDelegate_(BrowserView.DownloadDelegate.alloc().init().retain())

        def webView_decidePolicyForNavigationResponse_decisionHandler_(
            self, webview, navigationResponse, decisionHandler
        ):
            if navigationResponse.canShowMIMEType():
                response_ = navigationResponse.response()
                headers = dict(response_.allHeaderFields())
                response = Response(
                    response_.URL().absoluteString(), response_.statusCode(), headers
                )
                webview.pywebview_window.events.response_received.set(response)

                decisionHandler(WebKit.WKNavigationResponsePolicyAllow)
            elif webview_settings['ALLOW_DOWNLOADS']:
                decisionHandler(WebKit.WKNavigationResponsePolicyCancel)

                save_filename = navigationResponse.response().suggestedFilename()

                save_dlg = AppKit.NSSavePanel.savePanel()
                save_dlg.setTitle_(webview.pywebview_window.localization['global.saveFile'])
                directory = Foundation.NSSearchPathForDirectoriesInDomains(
                    Foundation.NSDownloadsDirectory, Foundation.NSUserDomainMask, True
                )[0]
                save_dlg.setDirectoryURL_(Foundation.NSURL.fileURLWithPath_(directory))

                if save_filename:  # set file name
                    save_dlg.setNameFieldStringValue_(save_filename)

                if save_dlg.runModal() == AppKit.NSFileHandlingPanelOKButton:
                    self._file_name = save_dlg.filename()
                    dataTask = Foundation.NSURLSession.sharedSession().downloadTaskWithURL_completionHandler_(
                        navigationResponse.response().URL(), self.download_completionHandler_error_
                    )
                    dataTask.resume()
                else:
                    self._file_name = None
            else:
                decisionHandler(WebKit.WKNavigationResponsePolicyCancel)

        def download_completionHandler_error_(self, temporaryLocation, response, error):
            if error is not None:
                logger.error('Download error:', error)
            else:
                if temporaryLocation is not None:
                    destinationURL = Foundation.NSURL.fileURLWithPath_(self._file_name)
                    fileManager = Foundation.NSFileManager.defaultManager()
                    try:
                        fileManager.moveItemAtURL_toURL_error_(
                            temporaryLocation, destinationURL, None
                        )
                    except Foundation.NSError as moveError:
                        logger.exception(moveError)

        # Show the webview when it finishes loading
        def webView_didFinishNavigation_(self, webview, nav):
            # Add the webview to the window if it's not yet the contentView
            i = BrowserView.get_instance('webview', webview)

            if i:
                if not webview.window():
                    i.window.setContentView_(webview)
                    i.window.makeFirstResponder_(webview)

                inject_pywebview('cocoa', i.js_bridge.window)

                if _state['debug'] and webview_settings['OPEN_DEVTOOLS_IN_DEBUG']:
                    BrowserView._open_web_inspector(webview)

        # Handle JavaScript window.print()
        def userContentController_didReceiveScriptMessage_(self, controller, message):
            if message.body() == 'print':
                i = BrowserView.get_instance('_browserDelegate', self)
                BrowserView.print_webview(i.webview)

    class FileFilterChooser(AppKit.NSPopUpButton):
        def initWithFilter_(self, file_filter):
            super(BrowserView.FileFilterChooser, self).init()
            self.filter = file_filter

            self.addItemsWithTitles_([i[0] for i in self.filter])
            self.setAction_('onChange:')
            self.setTarget_(self)
            return self

        def setFileDialog_(self, file_dlg):
            self.file_dlg = file_dlg

        def onChange_(self, sender):
            option = sender.indexOfSelectedItem()
            self.file_dlg.setAllowedFileTypes_(self.filter[option][1])

    class WebKitHost(WebKit.WKWebView):
        def performDragOperation_(self, sender):
            if sender.draggingSource() is None and _dnd_state['num_listeners'] > 0:
                pboard = sender.draggingPasteboard()
                classes = [AppKit.NSURL]
                options = {
                    AppKit.NSPasteboardURLReadingFileURLsOnlyKey: AppKit.NSNumber.numberWithBool_(
                        True
                    )
                }
                urls = pboard.readObjectsForClasses_options_(classes, options) or []
                files = [
                    (
                        os.path.basename(os.path.dirname(file_path))
                        if os.path.isdir(file_path)
                        else os.path.basename(file_path),
                        file_path,
                    )
                    for url in urls
                    for file_path in [
                        urllib.parse.unquote(
                            url.filePathURL().absoluteString().replace('file://', '')
                        )
                    ]
                    if os.path.isdir(file_path) or os.path.isfile(file_path)
                ]

                _dnd_state['paths'] += files

            return super(BrowserView.WebKitHost, self).performDragOperation_(sender)

        def addSubview_(self, event):
            # translate the event from the web coordinate space into the window coordinate space
            # basically flip the y axis since the WKWebView/WebKitHost implements a flipped view
            # (but seems to miss this flip back when adding a subview)
            event_frame = event.frame()
            flipped_y = self.frame().size.height - (event_frame.origin.y + event_frame.size.height)
            event.setFrameOrigin_(AppKit.NSPoint(event_frame.origin.x, flipped_y))

            super(BrowserView.WebKitHost, self).addSubview_(event)

        def mouseDown_(self, event):
            i = BrowserView.get_instance('webview', self)
            window = self.window()

            if i.frameless and i.easy_drag:
                windowFrame = window.frame()
                if windowFrame is None:
                    raise RuntimeError('Failed to obtain screen')

                self.initialLocation = window.convertBaseToScreen_(event.locationInWindow())
                self.initialLocation.x -= windowFrame.origin.x
                self.initialLocation.y -= windowFrame.origin.y

            super(BrowserView.WebKitHost, self).mouseDown_(event)

        def mouseDragged_(self, event):
            i = BrowserView.get_instance('webview', self)
            window = self.window()

            if i.frameless and i.easy_drag:
                screenFrame = i.screen
                if screenFrame is None:
                    raise RuntimeError('Failed to obtain screen')

                windowFrame = window.frame()
                if windowFrame is None:
                    raise RuntimeError('Failed to obtain frame')

                currentLocation = window.convertBaseToScreen_(
                    window.mouseLocationOutsideOfEventStream()
                )
                newOrigin = AppKit.NSMakePoint(
                    (currentLocation.x - self.initialLocation.x),
                    (currentLocation.y - self.initialLocation.y),
                )
                if (newOrigin.y + windowFrame.size.height) > (
                    screenFrame.origin.y + screenFrame.size.height
                ):
                    newOrigin.y = screenFrame.origin.y + (
                        screenFrame.size.height + windowFrame.size.height
                    )
                window.setFrameOrigin_(newOrigin)

            if event.modifierFlags() & getattr(AppKit, 'NSEventModifierFlagControl', 1 << 18):
                i = BrowserView.get_instance('webview', self)
                if not _state['debug']:
                    return

            super(BrowserView.WebKitHost, self).mouseDragged_(event)

        def willOpenMenu_withEvent_(self, menu, event):
            if not _state['debug']:
                menu.removeAllItems()

        def keyDown_(self, event):
            if event.modifierFlags() & AppKit.NSCommandKeyMask:
                responder = self.window().firstResponder()
                if responder is not None:
                    range_ = responder.selectedRange()
                    hasSelectedText = len(range_) > 0

                    char = event.characters()

                    if char == 'x' and hasSelectedText:  # cut
                        responder.cut_(self)
                        return
                    elif char == 'c' and hasSelectedText:  # copy
                        responder.copy_(self)
                        return
                    elif char == 'v':  # paste
                        responder.paste_(self)
                        return
                    elif char == 'a':  # select all
                        responder.selectAll_(self)
                        return
                    elif char == 'z':  # undo
                        if responder.undoManager().canUndo():
                            responder.undoManager().undo()
                        return
                    elif char == 'q':  # quit
                        BrowserView.app.stop_(self)
                        return
                    elif char == 'w':  # close
                        self.window().performClose_(event)
                        return

            super(BrowserView.WebKitHost, self).keyDown_(event)

    def __init__(self, window):
        BrowserView.instances[window.uid] = self
        self.uid = window.uid
        self.pywebview_window = window

        self.js_bridge = None
        self._file_name = None
        self._file_name_semaphore = Semaphore(0)
        self._current_url_semaphore = Semaphore(0)
        self.closed = window.events.closed
        self.closing = window.events.closing
        self.shown = window.events.shown
        self.loaded = window.events.loaded
        self.confirm_close = window.confirm_close
        self.title = window.title
        self.is_fullscreen = False
        self.hidden = window.hidden
        self.minimized = window.minimized
        self.maximized = window.maximized
        self.localization = window.localization

        if window.screen:
            self.screen = window.screen.frame
        else:
            self.screen = AppKit.NSScreen.mainScreen().frame()

        rect = AppKit.NSMakeRect(0.0, 0.0, window.initial_width, window.initial_height)
        window_mask = (
            AppKit.NSTitledWindowMask
            | AppKit.NSClosableWindowMask
            | AppKit.NSMiniaturizableWindowMask
        )

        if window.resizable:
            window_mask = window_mask | AppKit.NSResizableWindowMask

        if window.frameless:
            window_mask = (
                window_mask
                | NSFullSizeContentViewWindowMask
                | AppKit.NSTexturedBackgroundWindowMask
            )

        self.menu = window.menu or _state['menu']

        self.window = BrowserView.WindowHost.alloc().initWithContentRect_styleMask_backing_defer_(
            rect, window_mask, AppKit.NSBackingStoreBuffered, False
        )
        self.window.setReleasedWhenClosed_(False)
        self.pywebview_window.native = self.window

        self.window.focus = window.focus
        self.window.setTitle_(window.title)
        self.window.setMinSize_(AppKit.NSSize(window.min_size[0], window.min_size[1]))
        self.window.setAnimationBehavior_(AppKit.NSWindowAnimationBehaviorDocumentWindow)
        BrowserView.cascade_loc = self.window.cascadeTopLeftFromPoint_(BrowserView.cascade_loc)

        frame = self.window.frame()
        frame.size.width = window.initial_width
        frame.size.height = window.initial_height
        self.window.setFrame_display_(frame, True)

        config = WebKit.WKWebViewConfiguration.alloc().init()
        self.webview = BrowserView.WebKitHost.alloc().initWithFrame_configuration_(rect, config)
        self.webview.pywebview_window = window

        self._browserDelegate = BrowserView.BrowserDelegate.alloc().init()
        self._windowDelegate = BrowserView.WindowDelegate.alloc().init()

        if BrowserView._shared_app_delegate is None:
            BrowserView._shared_app_delegate = BrowserView.AppDelegate.alloc().init()
        BrowserView.app.setDelegate_(BrowserView._shared_app_delegate)
        self.webview.setUIDelegate_(self._browserDelegate)
        self.webview.setNavigationDelegate_(self._browserDelegate)
        self.window.setDelegate_(self._windowDelegate)

        config.userContentController().addScriptMessageHandler_name_(
            self._browserDelegate, 'browserDelegate'
        )

        self.datastore = WebKit.WKWebsiteDataStore.defaultDataStore()

        if _state['icon'] and os.path.isfile(_state['icon']):
            ns_image = AppKit.NSImage.alloc().initByReferencingFile_(_state['icon'])
            BrowserView.app.setApplicationIconImage_(ns_image)

        if _state['private_mode']:
            # nonPersisentDataStore preserves cookies for some unknown reason. For this reason we use default datastore
            # and clear all the cookies beforehand

            def dummy_completion_handler():
                pass

            data_types = WebKit.WKWebsiteDataStore.allWebsiteDataTypes()
            from_start = WebKit.NSDate.dateWithTimeIntervalSince1970_(0)
            config.setWebsiteDataStore_(self.datastore)
            self.datastore.removeDataOfTypes_modifiedSince_completionHandler_(
                data_types, from_start, dummy_completion_handler
            )
        else:
            config.setWebsiteDataStore_(self.datastore)

        try:
            config.preferences().setValue_forKey_(False, 'backspaceKeyNavigationEnabled')
        except KeyError:
            pass  # backspaceKeyNavigationEnabled does not exist prior to macOS Mojave

        config.preferences().setValue_forKey_(
            webview_settings['ALLOW_FILE_URLS'], 'allowFileAccessFromFileURLs'
        )

        if _state['debug']:
            config.preferences().setValue_forKey_(True, 'developerExtrasEnabled')

        self.js_bridge = BrowserView.JSBridge.alloc().initWithObject_(window)
        config.userContentController().addScriptMessageHandler_name_(self.js_bridge, 'jsBridge')

        user_agent = webview_settings.get('user_agent') or _state['user_agent']
        if user_agent:
            self.webview.setCustomUserAgent_(user_agent)

        self.window.setFrameOrigin_(self.screen.origin)

        if window.initial_x is not None and window.initial_y is not None:
            self.move(window.initial_x, window.initial_y)
        else:
            self.center()

        if window.transparent:
            self.window.setOpaque_(False)
            self.window.setHasShadow_(False)
            self.window.setBackgroundColor_(
                BrowserView.nscolor_from_hex(window.background_color, 0)
            )
            self.webview.setValue_forKey_(True, 'drawsTransparentBackground')
        else:
            self.window.setBackgroundColor_(BrowserView.nscolor_from_hex(window.background_color))

        if window.vibrancy:
            frame_vibrancy = AppKit.NSMakeRect(0, 0, frame.size.width, frame.size.height)
            visualEffectView = AppKit.NSVisualEffectView.new()
            visualEffectView.setAutoresizingMask_(
                AppKit.NSViewWidthSizable | AppKit.NSViewHeightSizable
            )
            visualEffectView.setWantsLayer_(True)
            visualEffectView.setFrame_(frame_vibrancy)
            visualEffectView.setState_(AppKit.NSVisualEffectStateActive)
            visualEffectView.setBlendingMode_(AppKit.NSVisualEffectBlendingModeBehindWindow)
            self.webview.addSubview_positioned_relativeTo_(
                visualEffectView, AppKit.NSWindowBelow, self.webview
            )

        self.frameless = window.frameless
        self.easy_drag = window.easy_drag

        if window.frameless:
            # Make content full size and titlebar transparent
            self.window.setTitlebarAppearsTransparent_(True)
            self.window.setTitleVisibility_(NSWindowTitleHidden)
            self.window.standardWindowButton_(AppKit.NSWindowCloseButton).setHidden_(True)
            self.window.standardWindowButton_(AppKit.NSWindowMiniaturizeButton).setHidden_(True)
            self.window.standardWindowButton_(AppKit.NSWindowZoomButton).setHidden_(True)
        else:
            # Set the titlebar color (so that it does not change with the window color)
            self.window.contentView().superview().subviews().lastObject().setBackgroundColor_(
                AppKit.NSColor.windowBackgroundColor()
            )

        if window.on_top:
            self.window.setLevel_(AppKit.NSStatusWindowLevel)

        self.pywebview_window.events.before_show.set()

        if window.real_url:
            self.url = window.real_url
            self.load_url(window.real_url)
        elif window.html:
            self.load_html(window.html, '')
        else:
            self.load_html(DEFAULT_HTML, '')
        if window.fullscreen:
            self.toggle_fullscreen()

    def first_show(self):
        if not self.hidden:
            self.window.makeKeyAndOrderFront_(self.window)

        if self.maximized:
            self.maximize()
        elif self.minimized:
            self.minimize()

        self.shown.set()

        if not BrowserView.app.isRunning():
            new_menu = self._recreate_menus(self.menu)
            BrowserView.app.setMainMenu_(new_menu)

            BrowserView.app.activateIgnoringOtherApps_(Foundation.YES)
            AppHelper.installMachInterrupt()
            BrowserView.app.run()

    def show(self):
        def _show():
            self.window.makeKeyAndOrderFront_(self.window)
            BrowserView.app.activateIgnoringOtherApps_(Foundation.YES)

        AppHelper.callAfter(_show)

    def hide(self):
        AppHelper.callAfter(self.window.orderOut_, self.window)

    def destroy(self):
        AppHelper.callAfter(self.window.close)

    def set_title(self, title):
        AppHelper.callAfter(self.window.setTitle_, title)

    def toggle_fullscreen(self):
        def toggle():
            if self.is_fullscreen:
                window_behaviour = 1 << 2  # NSWindowCollectionBehaviorManaged
            else:
                window_behaviour = 1 << 7  # NSWindowCollectionBehaviorFullScreenPrimary

            self.window.setCollectionBehavior_(window_behaviour)
            self.window.toggleFullScreen_(None)

        AppHelper.callAfter(toggle)
        self.is_fullscreen = not self.is_fullscreen

    def resize(self, width, height, fix_point):
        def _resize():
            frame = self.window.frame()

            if fix_point and fix_point & FixPoint.EAST:
                # Keep the right of the window in the same place
                frame.origin.x += frame.size.width - width

            if fix_point and fix_point & FixPoint.NORTH:
                # Keep the top of the window in the same place
                frame.origin.y += frame.size.height - height

            frame.size.width = width
            frame.size.height = height

            if not fix_point:  # used in maximized
                frame.origin.x = 0
                frame.origin.y = 0

            self.window.setFrame_display_(frame, True)

        AppHelper.callAfter(_resize)

    def maximize(self):
        width = self.screen.size.width
        height = self.screen.size.height
        self.resize(width, height, None)

    def minimize(self):
        AppHelper.callAfter(self.window.miniaturize_, self)

    def restore(self):
        AppHelper.callAfter(self.window.deminiaturize_, self)

    def move(self, x, y):
        flipped_y = self.screen.size.height - y
        self.window.setFrameTopLeftPoint_(
            AppKit.NSPoint(self.screen.origin.x + x, self.screen.origin.y + flipped_y)
        )

    def center(self):
        window_frame = self.window.frame()

        window_frame.origin.x = (
            self.screen.origin.x + (self.screen.size.width - window_frame.size.width) / 2
        )
        window_frame.origin.y = (
            self.screen.origin.y + (self.screen.size.height - window_frame.size.height) / 2
        )

        self.window.setFrameOrigin_(window_frame.origin)

    def clear_cookies(self):
        def clear():
            self.datastore.removeDataOfTypes_modifiedSince_completionHandler_(
                WebKit.WKWebsiteDataStore.allWebsiteDataTypes(),
                Foundation.NSDate.dateWithTimeIntervalSince1970_(0),
                lambda: None,
            )

        AppHelper.callAfter(clear)

    def get_cookies(self):
        def handler(cookies):
            for c in cookies:
                domain = c.domain()[1:] if c.domain().startswith('.') else c.domain()
                if domain not in self.url:
                    continue

                data = {
                    'name': c.name(),
                    'value': c.value(),
                    'path': c.path(),
                    'domain': c.domain(),
                    'expires': c.expiresDate(),
                    'secure': c.isSecure(),
                    'httponly': c.isHTTPOnly(),
                    'samesite': c.SameSitePolicy(),
                }

                cookie = create_cookie(data)
                _cookies.append(cookie)

            cookie_semaphore.release()

        _cookies = []
        AppHelper.callAfter(self.datastore.httpCookieStore().getAllCookies_, handler)
        cookie_semaphore = Semaphore(0)
        cookie_semaphore.acquire()

        return _cookies

    def get_current_url(self):
        def get():
            self._current_url = str(self.webview.URL())
            self._current_url_semaphore.release()

        AppHelper.callAfter(get)

        self._current_url_semaphore.acquire()
        return None if self._current_url == 'about:blank' else self._current_url

    def load_url(self, url):
        def load(url):
            page_url = Foundation.NSURL.URLWithString_(BrowserView.quote(url))
            req = Foundation.NSURLRequest.requestWithURL_(page_url)
            self.webview.loadRequest_(req)

        self.url = url
        AppHelper.callAfter(load, url)

    def load_html(self, content, base_uri):
        def load(content, url):
            url = Foundation.NSURL.URLWithString_(BrowserView.quote(url))
            self.webview.loadHTMLString_baseURL_(content, url)

        AppHelper.callAfter(load, content, base_uri)

    def evaluate_js(self, script, parse_json):
        def eval():
            self.webview.evaluateJavaScript_completionHandler_(script, handler)

        def handler(result, error):
            if parse_json and result:
                try:
                    JSResult.result = json.loads(result)
                except Exception:
                    logger.exception('Failed to parse JSON: ' + result)
                    JSResult.result = result
            else:
                JSResult.result = result

            JSResult.result_semaphore.release()

        class JSResult:
            result = None
            result_semaphore = Semaphore(0)

        AppHelper.callAfter(eval)
        JSResult.result_semaphore.acquire()
        return JSResult.result

    def create_file_dialog(
        self, dialog_type, directory, allow_multiple, save_filename, file_filter, main_thread=False
    ):
        def create_dialog(*args):
            dialog_type = args[0]

            if dialog_type == FileDialog.SAVE:
                save_filename = args[2]

                save_dlg = AppKit.NSSavePanel.savePanel()
                save_dlg.setTitle_(self.localization['global.saveFile'])

                if directory:  # set initial directory
                    save_dlg.setDirectoryURL_(Foundation.NSURL.fileURLWithPath_(directory))

                if save_filename:  # set file name
                    save_dlg.setNameFieldStringValue_(save_filename)

                if save_dlg.runModal() == AppKit.NSFileHandlingPanelOKButton:
                    self._file_name = save_dlg.filename()
                else:
                    self._file_name = None
            else:
                allow_multiple = args[1]

                open_dlg = AppKit.NSOpenPanel.openPanel()

                # Enable the selection of files in the dialog.
                open_dlg.setCanChooseFiles_(dialog_type != FileDialog.FOLDER)

                # Enable the selection of directories in the dialog.
                open_dlg.setCanChooseDirectories_(dialog_type == FileDialog.FOLDER)

                # Enable creating new folders in folder dialogs (macOS)
                if dialog_type == FileDialog.FOLDER:
                    open_dlg.setCanCreateDirectories_(True)

                # Enable / disable multiple selection
                open_dlg.setAllowsMultipleSelection_(allow_multiple)

                # Set allowed file extensions
                if file_filter:
                    if isinstance(file_filter, Foundation.WKNSArray):
                        try:
                            import UniformTypeIdentifiers

                            UTType = UniformTypeIdentifiers.UTType
                        except ImportError:
                            UTType = None  # Fallback if UTType is not available

                        if UTType is not None:
                            # Map wildcard MIME types to supertype UTIs because
                            # typeWithMIMEType_ returns unusable dynamic types for them.
                            _wildcard_uti = {
                                'image/*': 'public.image',
                                'video/*': 'public.movie',
                                'audio/*': 'public.audio',
                                'text/*': 'public.text',
                            }
                            content_types = []
                            for mime in file_filter:
                                mime_str = str(mime)
                                uti_id = _wildcard_uti.get(mime_str)
                                if uti_id:
                                    ut = UTType.typeWithIdentifier_(uti_id)
                                else:
                                    ut = UTType.typeWithMIMEType_(mime_str)
                                if ut is not None and not str(ut).startswith('dyn.'):
                                    content_types.append(ut)
                            if content_types:
                                open_dlg.setAllowedContentTypes_(content_types)
                    else:
                        open_dlg.setAllowedFileTypes_(file_filter[0][1])

                        if len(file_filter) > 1:
                            filter_chooser = BrowserView.FileFilterChooser.alloc().initWithFilter_(
                                file_filter
                            )
                            filter_chooser.setFileDialog_(open_dlg)
                            open_dlg.setAccessoryView_(filter_chooser)
                            open_dlg.setAccessoryViewDisclosed_(True)

                if directory:  # set initial directory
                    open_dlg.setDirectoryURL_(Foundation.NSURL.fileURLWithPath_(directory))

                if open_dlg.runModal() == AppKit.NSFileHandlingPanelOKButton:
                    files = open_dlg.filenames()
                    self._file_name = tuple(files)
                else:
                    self._file_name = None

            if not main_thread:
                self._file_name_semaphore.release()

        if main_thread:
            create_dialog(dialog_type, allow_multiple, save_filename)
        else:
            AppHelper.callAfter(create_dialog, dialog_type, allow_multiple, save_filename)
            self._file_name_semaphore.acquire()

        return self._file_name

    def _recreate_menus(self, user_menu):
        main_menu = self._clear_main_menu()

        # Filter out app menu items (menus with title '__app__')
        app_menu_items = None
        regular_menus = None

        if user_menu:
            app_menu_items = []
            regular_menus = []

            for menu in user_menu:
                if isinstance(menu, Menu) and menu.title == '__app__':
                    # This is an app menu - extract its items
                    app_menu_items.extend(menu.items)
                else:
                    # Regular menu
                    regular_menus.append(menu)

            # If no app menu items were found, set to None
            if not app_menu_items:
                app_menu_items = None

            # If no regular menus, set to None
            if not regular_menus:
                regular_menus = None

        self._add_app_menu(main_menu, app_menu_items)

        if webview_settings['SHOW_DEFAULT_MENUS']:
            self._add_view_menu(main_menu)
            self._add_edit_menu(main_menu)

        self._add_custom_menu(main_menu, regular_menus)

        return main_menu

    def _clear_main_menu(self):
        """
        Remove all items from the main menu or create a new one if it doesn't exist.
        """
        mainMenu = BrowserView.app.mainMenu()
        if mainMenu:
            mainMenu.removeAllItems()

        mainMenu = AppKit.NSMenu.alloc().init()

        return mainMenu

    def _add_app_menu(self, mainMenu, custom_items=None):
        """
        Create a default Cocoa menu that shows 'Services', 'Hide',
        'Hide Others', 'Show All', and 'Quit'. Will append the application name
        to some menu items if it's available.
        """

        # Create an application menu and make it a submenu of the main menu
        mainAppMenuItem = AppKit.NSMenuItem.alloc().init()
        # The application menu is the first item, so add this menu ast the first item
        mainMenu.insertItem_atIndex_(mainAppMenuItem, 0)
        appMenu = AppKit.NSMenu.alloc().init()
        mainAppMenuItem.setSubmenu_(appMenu)

        appMenu.addItemWithTitle_action_keyEquivalent_(
            self._append_app_name(self.localization['cocoa.menu.about']),
            'orderFrontStandardAboutPanel:',
            '',
        )

        # Add custom app menu items if provided (between About and Services)
        if custom_items:
            appMenu.addItem_(AppKit.NSMenuItem.separatorItem())
            self._process_menu_items(custom_items, appMenu)

        appMenu.addItem_(AppKit.NSMenuItem.separatorItem())

        # Set the 'Services' menu for the app and create an app menu item
        appServicesMenu = AppKit.NSMenu.alloc().init()
        BrowserView.app.setServicesMenu_(appServicesMenu)
        servicesMenuItem = appMenu.addItemWithTitle_action_keyEquivalent_(
            self.localization['cocoa.menu.services'], nil, ''
        )
        servicesMenuItem.setSubmenu_(appServicesMenu)

        appMenu.addItem_(AppKit.NSMenuItem.separatorItem())

        # Append the 'Hide', 'Hide Others', and 'Show All' menu items
        appMenu.addItemWithTitle_action_keyEquivalent_(
            self._append_app_name(self.localization['cocoa.menu.hide']), 'hide:', 'h'
        )
        hideOthersMenuItem = appMenu.addItemWithTitle_action_keyEquivalent_(
            self.localization['cocoa.menu.hideOthers'], 'hideOtherApplications:', 'h'
        )
        hideOthersMenuItem.setKeyEquivalentModifierMask_(
            AppKit.NSAlternateKeyMask | AppKit.NSCommandKeyMask
        )
        appMenu.addItemWithTitle_action_keyEquivalent_(
            self.localization['cocoa.menu.showAll'], 'unhideAllApplications:', ''
        )

        appMenu.addItem_(AppKit.NSMenuItem.separatorItem())

        # Append a 'Quit' menu item
        appMenu.addItemWithTitle_action_keyEquivalent_(
            self._append_app_name(self.localization['cocoa.menu.quit']), 'terminate:', 'q'
        )

    def _add_view_menu(self, mainMenu):
        """
        Create a default View menu that shows 'Enter Full Screen'.
        """

        # Create an View menu and make it a submenu of the main menu
        viewMenu = AppKit.NSMenu.alloc().init()
        viewMenu.setTitle_(self.localization['cocoa.menu.view'])
        viewMenuItem = AppKit.NSMenuItem.alloc().init()
        viewMenuItem.setSubmenu_(viewMenu)
        # Make the view menu the first item after the application menu
        mainMenu.insertItem_atIndex_(viewMenuItem, 1)

        # TODO: localization of the Enter fullscreen string has no effect
        fullScreenMenuItem = viewMenu.addItemWithTitle_action_keyEquivalent_(
            self.localization['cocoa.menu.fullscreen'], 'toggleFullScreen:', 'f'
        )
        fullScreenMenuItem.setKeyEquivalentModifierMask_(
            AppKit.NSControlKeyMask | AppKit.NSCommandKeyMask
        )

    def _add_edit_menu(self, mainMenu):
        """
        Create a default Edit menu that shows Copy/Paste/etc.
        """
        # Create an Edit menu and make it a submenu of the main menu
        editMenu = AppKit.NSMenu.alloc().init()
        editMenu.setTitle_(self.localization['cocoa.menu.edit'])
        editMenuItem = AppKit.NSMenuItem.alloc().init()
        editMenuItem.setSubmenu_(editMenu)
        # Make the edit menu the first item after the application menu
        mainMenu.insertItem_atIndex_(editMenuItem, 1)

        for title, action, keyEquivalent in [
            (self.localization['cocoa.menu.cut'], 'cut:', 'x'),
            (self.localization['cocoa.menu.copy'], 'copy:', 'c'),
            (self.localization['cocoa.menu.paste'], 'paste:', 'v'),
            (self.localization['cocoa.menu.selectAll'], 'selectAll:', 'a'),
        ]:
            editMenu.addItemWithTitle_action_keyEquivalent_(title, action, keyEquivalent)

    def _process_menu_items(self, menu_items, parent_menu):
        """
        Process menu items and add them to the parent menu.
        Used for both custom menus and app menu items.
        """
        for item in menu_items:
            if isinstance(item, MenuSeparator):
                parent_menu.addItem_(AppKit.NSMenuItem.separatorItem())
            elif isinstance(item, MenuAction):
                # Actions must be registered before application start. Otherwise they are disabled.
                # Menu handler is a workaround to register actions after application start
                random_id = str(uuid.uuid4())[:6]
                # Handle functools.partial objects which don't have __name__ attribute
                if hasattr(item.function, '__name__'):
                    func_name = item.function.__name__
                elif hasattr(item.function, 'func') and hasattr(item.function.func, '__name__'):
                    func_name = item.function.func.__name__
                else:
                    func_name = 'anonymous_function'
                action_id = func_name + '.' + random_id
                menu_handler.register_action(action_id, item.function)

                menu_item = AppKit.NSMenuItem.alloc().initWithTitle_action_keyEquivalent_(
                    item.title, 'handleMenuAction:', ''
                )
                menu_item.setTarget_(menu_handler)
                menu_item.setRepresentedObject_(action_id)
                parent_menu.addItem_(menu_item)
            elif isinstance(item, Menu):
                submenu = AppKit.NSMenu.alloc().init()
                submenu.setTitle_(item.title)
                menu_item = AppKit.NSMenuItem.alloc().init()
                menu_item.setTitle_(item.title)
                menu_item.setSubmenu_(submenu)
                parent_menu.addItem_(menu_item)

                self._process_menu_items(item.items, submenu)

    def _add_custom_menu(self, mainMenu, app_menu_list):
        """
        Create a custom menu for the app menu (MacOS bar menu)
        """

        if app_menu_list is None:
            return

        for app_menu in app_menu_list:
            submenu = AppKit.NSMenu.alloc().init()
            submenu.setTitle_(app_menu.title)
            menu_item = AppKit.NSMenuItem.alloc().init()
            menu_item.setTitle_(app_menu.title)
            menu_item.setSubmenu_(submenu)
            mainMenu.addItem_(menu_item)
            self._process_menu_items(app_menu.items, submenu)

    def _append_app_name(self, val):
        """
        Append the application name to a string if it's available. If not, the
        string is returned unchanged.

        :param str val: The string to append to
        :return: String with app name appended, or unchanged string
        :rtype: str
        """
        if 'CFBundleName' in info:
            val += ' {}'.format(info['CFBundleName'])
        return val

    @staticmethod
    def nscolor_from_hex(hex_string, alpha=1.0):
        """
        Convert given hex color to NSColor.

        :hex_string: Hex code of the color as #RGB or #RRGGBB
        """

        hex_string = hex_string[1:]  # Remove leading hash
        if len(hex_string) == 3:
            hex_string = ''.join([c * 2 for c in hex_string])  # 3-digit to 6-digit

        hex_int = int(hex_string, 16)
        rgb = (
            (hex_int >> 16) & 0xFF,  # Red byte
            (hex_int >> 8) & 0xFF,  # Blue byte
            (hex_int) & 0xFF,  # Green byte
        )
        rgb = [i / 255.0 for i in rgb]  # Normalize to range(0.0, 1.0)

        return AppKit.NSColor.colorWithSRGBRed_green_blue_alpha_(rgb[0], rgb[1], rgb[2], alpha)

    @staticmethod
    def get_instance(attr, value):
        """
        Return a BrowserView instance by the :value of its given :attribute,
        and None if no match is found.
        """
        for i in list(BrowserView.instances.values()):
            try:
                if getattr(i, attr) == value:
                    return i
            except AttributeError:
                break

        return None

    @staticmethod
    def display_confirmation_dialog(first_button, second_button, message):
        AppKit.NSApplication.sharedApplication()
        AppKit.NSRunningApplication.currentApplication().activateWithOptions_(
            AppKit.NSApplicationActivateIgnoringOtherApps
        )
        alert = AppKit.NSAlert.alloc().init()
        alert.addButtonWithTitle_(first_button)
        alert.addButtonWithTitle_(second_button)
        alert.setMessageText_(message)
        alert.setAlertStyle_(AppKit.NSWarningAlertStyle)

        return alert.runModal() == AppKit.NSAlertFirstButtonReturn

    @staticmethod
    def display_input_dialog(first_button, second_button, prompt, default_text):
        AppKit.NSApplication.sharedApplication()
        AppKit.NSRunningApplication.currentApplication().activateWithOptions_(
            AppKit.NSApplicationActivateIgnoringOtherApps
        )
        alert = AppKit.NSAlert.alloc().init()
        text_field = AppKit.NSTextField.alloc().initWithFrame_(AppKit.NSMakeRect(0, 0, 240, 24))
        text_field.cell().setScrollable_(True)
        text_field.setStringValue_(default_text)
        alert.setAccessoryView_(text_field)
        alert.addButtonWithTitle_(first_button)
        alert.addButtonWithTitle_(second_button)
        alert.setMessageText_(prompt)
        alert.setAlertStyle_(AppKit.NSWarningAlertStyle)

        if alert.runModal() == AppKit.NSAlertFirstButtonReturn:
            return text_field.stringValue()
        else:
            return None

    @staticmethod
    def should_close(window):
        quit = window.localization['global.quit']
        cancel = window.localization['global.cancel']
        msg = window.localization['global.quitConfirmation']

        should_cancel = window.events.closing.set()
        if should_cancel:
            return Foundation.NO
        if not window.confirm_close or BrowserView.display_confirmation_dialog(quit, cancel, msg):
            return Foundation.YES
        return Foundation.NO

    @staticmethod
    def print_webview(webview):
        info = AppKit.NSPrintInfo.sharedPrintInfo().copy()

        # default print settings used by Safari
        info.setHorizontalPagination_(AppKit.NSFitPagination)
        info.setHorizontallyCentered_(Foundation.NO)
        info.setVerticallyCentered_(Foundation.NO)

        imageableBounds = info.imageablePageBounds()
        paperSize = info.paperSize()
        if Foundation.NSWidth(imageableBounds) > paperSize.width:
            imageableBounds.origin.x = 0
            imageableBounds.size.width = paperSize.width
        if Foundation.NSHeight(imageableBounds) > paperSize.height:
            imageableBounds.origin.y = 0
            imageableBounds.size.height = paperSize.height

        info.setBottomMargin_(Foundation.NSMinY(imageableBounds))
        info.setTopMargin_(
            paperSize.height
            - Foundation.NSMinY(imageableBounds)
            - Foundation.NSHeight(imageableBounds)
        )
        info.setLeftMargin_(Foundation.NSMinX(imageableBounds))
        info.setRightMargin_(
            paperSize.width
            - Foundation.NSMinX(imageableBounds)
            - Foundation.NSWidth(imageableBounds)
        )

        # show the print panel
        print_op = webview._printOperationWithPrintInfo_(info)
        print_op.runOperationModalForWindow_delegate_didRunSelector_contextInfo_(
            webview.window(), nil, nil, nil
        )

    @staticmethod
    def _open_web_inspector(webview):
        """
        Programmatically open the Web Inspector for the given WKWebView.
        Uses private WebKit APIs that may not work on all macOS versions.
        """
        try:
            if hasattr(webview, '_inspector'):
                inspector = webview._inspector()
                if inspector and hasattr(inspector, 'show'):
                    inspector.show()
                    return True

            return False

        except Exception:
            return False

    @staticmethod
    def quote(string):
        return string.replace(' ', '%20')


class MenuHandler:
    def __init__(self):
        self.actions = {}

    def handleMenuAction_(self, sender):
        action_id = sender.representedObject()
        if action_id in self.actions:
            Thread(target=self.actions[action_id]).start()
        else:
            logger.warning(f'Menu cction {action_id} not found')

    def register_action(self, action_id, action_callable):
        self.actions[action_id] = action_callable


menu_handler = MenuHandler()


def setup_app():
    pass


def get_active_window():
    active_window = BrowserView.app.keyWindow()
    if active_window is None:
        return None

    active_window_number = active_window.windowNumber()

    for uid, browser_view_instance in BrowserView.instances.items():
        if browser_view_instance.window.windowNumber() == active_window_number:
            return browser_view_instance.pywebview_window

    return None


def create_window(window):
    def create():
        browser = BrowserView(window)
        browser.first_show()

    if window.uid == 'master':
        main_thread().pydev_do_not_trace = True  # vs code debugger hang fix
        create()

    else:
        AppHelper.callAfter(create)


def set_title(title, uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.set_title(title)


def create_confirmation_dialog(title, message, uid):
    def _confirm():
        nonlocal result

        i = BrowserView.instances.get(uid)
        ok = i.localization['global.ok']
        cancel = i.localization['global.cancel']

        result = BrowserView.display_confirmation_dialog(ok, cancel, message)
        semaphore.release()

    result = False

    semaphore = Semaphore(0)
    AppHelper.callAfter(_confirm)
    semaphore.acquire()

    return result


def create_file_dialog(dialog_type, directory, allow_multiple, save_filename, file_types, uid):
    file_filter = []

    # Parse file_types to obtain allowed file extensions
    for s in file_types:
        description, extensions = parse_file_type(s)
        file_extensions = [i.lstrip('*.') for i in extensions.split(';') if i != '*.*']
        file_filter.append([description, file_extensions or []])

    i = BrowserView.instances.get(uid)
    return i.create_file_dialog(dialog_type, directory, allow_multiple, save_filename, file_filter)


def load_url(url, uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.load_url(url)


def load_html(content, base_uri, uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.load_html(content, base_uri)


def destroy_window(uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.destroy()


def hide(uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.hide()


def show(uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.show()


def toggle_fullscreen(uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.toggle_fullscreen()


def set_on_top(uid, top):
    def _set_on_top():
        level = AppKit.NSStatusWindowLevel if top else AppKit.NSNormalWindowLevel
        i.window.setLevel_(level)

    i = BrowserView.instances.get(uid)
    if i:
        AppHelper.callAfter(_set_on_top)


def resize(width, height, uid, fix_point):
    i = BrowserView.instances.get(uid)
    if i:
        i.resize(width, height, fix_point)


def maximize(uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.maximize()


def minimize(uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.minimize()


def restore(uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.restore()


def move(x, y, uid):
    i = BrowserView.instances.get(uid)
    if i:
        AppHelper.callAfter(i.move, x, y)


def get_current_url(uid):
    i = BrowserView.instances.get(uid)
    if i:
        return i.get_current_url()


def clear_cookies(uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.clear_cookies()


def get_cookies(uid):
    i = BrowserView.instances.get(uid)
    if i:
        return i.get_cookies()


def evaluate_js(script, uid, parse_json=True):
    i = BrowserView.instances.get(uid)
    if i:
        return i.evaluate_js(script, parse_json)


def get_position(uid):
    def _position(coordinates):
        screen_frame = i.screen

        if screen_frame is None:
            raise RuntimeError('Failed to obtain screen')

        window = i.window
        frame = window.frame()
        coordinates[0] = int(frame.origin.x)
        coordinates[1] = int(screen_frame.size.height - frame.origin.y - frame.size.height)
        semaphore.release()

    coordinates = [None, None]
    semaphore = Semaphore(0)

    i = BrowserView.instances.get(uid)
    if not i:
        return None, None

    try:
        _position(coordinates)
    except Exception:
        AppHelper.callAfter(_position, coordinates)
        semaphore.acquire()

    return coordinates


def get_size(uid):
    def _size(dimensions):
        size = i.window.frame().size
        dimensions[0] = size.width
        dimensions[1] = size.height
        semaphore.release()

    dimensions = [None, None]
    semaphore = Semaphore(0)

    i = BrowserView.instances.get(uid)
    if not i:
        return None, None

    try:
        _size(dimensions)
    except Exception:
        AppHelper.callAfter(_size, dimensions)
        semaphore.acquire()

    return dimensions


def get_screens():
    screens = [
        Screen(
            s.frame().origin.x,
            s.frame().origin.y,
            s.frame().size.width,
            s.frame().size.height,
            s.frame(),
            s.backingScaleFactor(),
        )
        for s in AppKit.NSScreen.screens()
    ]
    return screens


def add_tls_cert(certfile):
    # does not auth against the certfile
    # see webView_didReceiveAuthenticationChallenge_completionHandler_
    pass



import logging
from collections import defaultdict
from functools import wraps
from typing import Any, Callable, Dict, Iterable, List, Optional, Union

from webview.dom import DOMEventHandler, ManipulationMode, _dnd_state
from webview.dom.classlist import ClassList
from webview.dom.propsdict import DOMPropType, PropsDict
from webview.errors import JavascriptException
from webview.event import EventContainer

logger = logging.getLogger('pywebview')


def _ignore_window_document(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if args[0]._node_id in ('window', 'document'):
            return None

        return func(*args, **kwargs)

    return wrapper


def _exists(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if not args[0]._exists:
            logger.warning(f'Element with id {args[0]._node_id} has been removed')
            return

        try:
            return func(*args, **kwargs)
        except JavascriptException as e:
            if e.args[0].get('cause') == 'ELEMENT_NOT_FOUND':
                logger.warning(f'Element with id {args[0]._node_id} has been removed')
            else:
                logger.exception(e)
            return

    return wrapper


class Element:
    def __init__(self, window, node_id) -> None:
        self._window = window
        self.events = EventContainer()
        self._node_id = node_id
        self._query_command = rf"""
            var element;

            if ('{self._node_id}' === 'document') {{
                element = document;
            }} else if ('{self._node_id}' === 'window') {{
                element = window;
            }} else if ('{self._node_id}' === 'body') {{
                element = document.body;
            }} else {{
                element = document.querySelector('[data-pywebview-id="{self._node_id}"]');
            }}

            if (!element) {{
                throw new Error('Element with pywebview-id {self._node_id} not found', {{ cause: 'ELEMENT_NOT_FOUND' }});
            }}
        """.replace('\n', '')
        self._event_handlers = defaultdict(list)
        self._event_handler_ids = {}
        self._exists = True
        self._classes = ClassList(self)
        self._style = PropsDict(self, DOMPropType.Style)
        self._attributes = PropsDict(self, DOMPropType.Attribute)

        self.__original_display = None
        self.__generate_events()

    @property
    @_exists
    @_ignore_window_document
    def tag(self) -> str:
        return self._window.evaluate_js(f'{self._query_command}; element.tagName').lower()

    @property
    @_exists
    @_ignore_window_document
    def id(self) -> Optional[str]:
        return self._window.evaluate_js(f'{self._query_command}; element.id')

    @id.setter
    @_exists
    @_ignore_window_document
    def id(self, id: str) -> None:
        self._window.run_js(f"{self._query_command}; element.id = '{id}'")

    @property
    @_exists
    @_ignore_window_document
    def classes(self) -> ClassList:
        return self._classes

    @classes.setter
    def classes(self, classes: Iterable) -> None:
        self._classes = ClassList(self, classes)

    @property
    @_exists
    @_ignore_window_document
    def attributes(self) -> Dict[str, Any]:
        return self._attributes

    @attributes.setter
    @_exists
    @_ignore_window_document
    def attributes(self, attributes: Dict[str, Any]) -> None:
        self._attributes = PropsDict(self, DOMPropType.Attribute, attributes)

    @property
    @_exists
    def node(self) -> Dict[str, Any]:
        return self._window.evaluate_js(
            f'{self._query_command}; var r2 = pywebview._processElements([element])[0]; r2'
        )

    @property
    @_exists
    @_ignore_window_document
    def style(self) -> Dict[str, Any]:
        return self._style

    @style.setter
    @_exists
    @_ignore_window_document
    def style(self, style: Dict[str, Any]) -> None:
        self._style = PropsDict(self, DOMPropType.Style, style)

    @property
    @_exists
    @_ignore_window_document
    def tabindex(self) -> int:
        return self._window.evaluate_js(f'{self._query_command}; element.tabIndex')

    @tabindex.setter
    @_exists
    @_ignore_window_document
    def tabindex(self, tab_index: int) -> None:
        self._window.run_js(f'{self._query_command}; element.tabIndex = {tab_index}')

    @property
    @_exists
    @_ignore_window_document
    def text(self) -> str:
        return self._window.evaluate_js(f'{self._query_command}; element.textContent')

    @text.setter
    @_exists
    @_ignore_window_document
    def text(self, text: str) -> None:
        self._window.run_js(f"{self._query_command}; element.textContent = '{text}'")

    @property
    @_exists
    @_ignore_window_document
    def value(self) -> str:
        return self._window.evaluate_js(f'{self._query_command}; element.value')

    @property
    @_exists
    @_ignore_window_document
    def visible(self) -> bool:
        return self._window.evaluate_js(f'{self._query_command}; element.offsetParent !== null')

    @property
    @_exists
    @_ignore_window_document
    def focused(self) -> bool:
        return self._window.evaluate_js(
            f'{self._query_command}; document.activeElement === element'
        )

    @value.setter
    @_exists
    @_ignore_window_document
    def value(self, value: str) -> None:
        self._window.run_js(
            f"{self._query_command}; if ('value' in element) {{ element.value = '{value}' }}"
        )

    @_exists
    @_ignore_window_document
    def blur(self) -> None:
        self._window.run_js(f'{self._query_command}; element.blur()')

    @_exists
    @_ignore_window_document
    def focus(self) -> None:
        self._window.run_js(f'{self._query_command}; element.focus()')

    @property
    @_exists
    @_ignore_window_document
    def children(self) -> List['Element']:
        children = self._window.evaluate_js(
            f"""
            {self._query_command};
            var children = element.children;
            var nodeIds = []
            for (var i = 0; i < children.length; i++) {{
                var nodeId = pywebview._getNodeId(children[i]);
                nodeIds.push(nodeId);
            }}

            nodeIds
        """
        )
        return [Element(self._window, node_id) for node_id in children]

    @property
    @_exists
    @_ignore_window_document
    def parent(self) -> Union['Element', None]:
        node_id = self._window.evaluate_js(
            f"""
            {self._query_command};
            var parent = element.parentElement;
            parent ? pywebview._getNodeId(parent) : null
        """
        )
        return Element(self._window, node_id) if node_id else None

    @property
    @_exists
    @_ignore_window_document
    def next(self) -> Union['Element', None]:
        node_id = self._window.evaluate_js(
            f"""
            {self._query_command};
            var nextSibling = element.nextElementSibling;
            nextSibling ? pywebview._getNodeId(nextSibling) : null
        """
        )
        return Element(self._window, node_id) if node_id else None

    @property
    @_exists
    @_ignore_window_document
    def previous(self) -> Union['Element', None]:
        node_id = self._window.evaluate_js(
            f"""
            {self._query_command};
            var previousSibling = element.previousElementSibling;
            previousSibling ? pywebview._getNodeId(previousSibling) : null
        """
        )
        return Element(self._window, node_id) if node_id else None

    @_exists
    @_ignore_window_document
    def hide(self) -> None:
        self.__original_display = self._window.evaluate_js(
            f'{self._query_command}; element.style.display'
        )
        self._window.run_js(f"{self._query_command}; element.style.display = 'none'")

    @_exists
    @_ignore_window_document
    def show(self) -> None:
        current_display = self._window.evaluate_js(f'{self._query_command}; element.style.display')

        if current_display == 'none':
            display = self.__original_display or 'block'
            self._window.run_js(f"{self._query_command}; element.style.display = '{display}'")

    @_exists
    @_ignore_window_document
    def toggle(self) -> None:
        current_display = self._window.evaluate_js(f'{self._query_command}; element.style.display')

        if current_display == 'none':
            self.show()
        else:
            self.hide()

    @_exists
    @_ignore_window_document
    def append(self, html: str, mode=ManipulationMode.LastChild) -> 'Element':
        return self._window.dom.create_element(html, self, mode)

    @_exists
    @_ignore_window_document
    def empty(self) -> None:
        self._window.run_js(f"{self._query_command}; element.innerHTML = ''")

    @_exists
    @_ignore_window_document
    def remove(self) -> None:
        self._window.run_js(f'{self._query_command}; element.remove()')

        if self._node_id in self._window.dom._elements:
            self._window.dom._elements.pop(self._node_id)

        handler_ids = ','.join(
            [f'"{handler_id}"' for handler_id in self._event_handler_ids.values()]
        )
        self._window.run_js(
            f"""
            var handlerIds = [{handler_ids}];
            handlerIds.forEach(function(handlerId) {{
                delete pywebview._eventHandlers[handler_id]
            }})
        """
        )
        self._event_handler_ids = {}
        self._event_handlers = defaultdict(list)
        self._exists = False

    @_exists
    @_ignore_window_document
    def copy(
        self, target: Union[str, 'Element'] = None, mode=ManipulationMode.LastChild, id: str = None
    ) -> 'Element':
        if isinstance(target, str):
            target = self._window.dom.get_element(target)
        elif target is None:
            target = self.parent

        if id:
            id_command = f'newElement.id = "{id}"'
        else:
            id_command = 'newElement.removeAttribute("id")'

        node_id = self._window.evaluate_js(
            f"""
            {self._query_command};
            var target = document.querySelector('[data-pywebview-id=\"{target._node_id}\"]');
            var newElement = element.cloneNode(true);
            newElement.removeAttribute('data-pywebview-id');
            {id_command};

            var nodeId = pywebview._getNodeId(newElement);
            pywebview._insertNode(newElement, target, '{mode.value}')
            nodeId;
        """
        )

        new_element = Element(self._window, node_id)
        for event, handlers in self._event_handlers.items():
            for handler in handlers:
                new_element.on(event, handler)

        return new_element

    @_exists
    @_ignore_window_document
    def move(self, target: Union[str, 'Element'], mode=ManipulationMode.LastChild) -> 'Element':
        if isinstance(target, str):
            target = self._window.dom.get_element(target)

        self._window.run_js(
            f"""
            {self._query_command};
            var target = document.querySelector('[data-pywebview-id=\"{target._node_id}\"]');
            pywebview._insertNode(element, target, '{mode.value}')
        """
        )
        return self

    @_exists
    def on(self, event: str, callback: Union[Callable, DOMEventHandler]) -> None:
        if self._node_id not in self._window.dom._elements:
            self._window.dom._elements[self._node_id] = self

        if isinstance(callback, DOMEventHandler):
            prevent_default = 'e.preventDefault();' if callback.prevent_default else ''
            stop_propagation = 'e.stopPropagation();' if callback.stop_propagation else ''
            debounce = callback.debounce
            callback = callback.callback
        else:
            prevent_default = ''
            stop_propagation = ''
            debounce = 0

        callback_func = f"window.pywebview._jsApiCallback('pywebviewEventHandler', {{ event: e, nodeId: '{self._node_id}' }}, 'eventHandler')"
        debounced_func = (
            f'pywebview._debounce(function() {{ {callback_func} }}, {debounce})'
            if debounce > 0
            else callback_func
        )

        handler_id = self._window.evaluate_js(
            f"""
            {self._query_command};
            var handlerId = null

            if (element) {{
                handlerId = Math.random().toString(36).substr(2, 11);
                pywebview._eventHandlers[handlerId] = function(e) {{
                    {prevent_default}
                    {stop_propagation}
                    {debounced_func};
                }}

                element.addEventListener('{event}', pywebview._eventHandlers[handlerId]);
            }};
            handlerId;
        """
        )

        if handler_id:
            self._event_handlers[event].append(callback)
            self._event_handler_ids[callback] = handler_id

        if event == 'drop':
            _dnd_state['num_listeners'] += 1

    @_exists
    def off(self, event: str, callback: Callable) -> None:
        handler_id = self._event_handler_ids.get(callback)

        if not handler_id:
            return

        self._window.run_js(
            f"""
            {self._query_command};
            var callback = pywebview._eventHandlers['{handler_id}'];
            if (element) {{
                element.removeEventListener('{event}', callback);
                delete pywebview._eventHandlers['{handler_id}'];
            }}
        """
        )

        del self._event_handler_ids[callback]
        del self._window.dom._elements[self._node_id]

        for handler in self._event_handlers[event]:
            if handler == callback:
                self._event_handlers[event].remove(handler)

        if event == 'drop':
            _dnd_state['num_listeners'] = max(0, _dnd_state['num_listeners'] - 1)

    @_exists
    def __generate_events(self):
        from webview.dom.event import DOMEvent

        events = self._window.evaluate_js(
            f"""
            {self._query_command};
            Object
                .getOwnPropertyNames(element)
                .concat(
                    Object.getOwnPropertyNames(
                        Object.getPrototypeOf(
                            Object.getPrototypeOf(element)
                        )
                    )
                )
                .filter(function(i){{
                    return !i.indexOf('on') && (element[i] == null || typeof element[i]=='function');
                }})
                .map(function(f) {{ return f.substr(2) }});
        """
        )

        for event in events:
            setattr(self.events, event, DOMEvent(event, self))

    @_exists
    def __str__(self) -> str:
        return repr(self)

    @_exists
    def __repr__(self) -> str:
        if self.node['nodeName'] == '#document':
            return 'document'
        elif self.node['nodeName'] == '#window':
            return 'window'
        else:
            return self.node['outerHTML']

    @_exists
    def __eq__(self, other: 'Element') -> bool:
        return hasattr(other, '_node_id') and self._node_id == other._node_id



__all__ = ('View', 'KeyEvent', 'Choreographer')

from jnius import JavaClass, JavaMethod, JavaStaticField, JavaStaticMethod, MetaJavaClass


class View(JavaClass, metaclass=MetaJavaClass):
    """
    Represents an Android view class with constants for system UI visibility flags.

    Provides direct references to various system UI visibility constants, which
    are used to control the appearance of the system user interface elements
    such as the status bar, navigation bar, and fullscreen modes. This class
    serves as a bridge to work with Java's Android View API in Python through
    a metaclass-based approach.

    Attributes
    ----------
    __javaclass__ : str
        The name of the corresponding Java class used for this bridge.
    SYSTEM_UI_FLAG_FULLSCREEN : JavaStaticField
        Constant for enabling fullscreen mode in the system UI.
    SYSTEM_UI_FLAG_HIDE_NAVIGATION : JavaStaticField
        Constant for hiding the navigation bar in the system UI.
    SYSTEM_UI_FLAG_IMMERSIVE : JavaStaticField
        Constant for enabling immersive mode.
    SYSTEM_UI_FLAG_IMMERSIVE_STICKY : JavaStaticField
        Constant for enabling sticky immersive mode.
    SYSTEM_UI_FLAG_LAYOUT_FULLSCREEN : JavaStaticField
        Constant for making the layout occupy the fullscreen area.
    SYSTEM_UI_FLAG_LAYOUT_HIDE_NAVIGATION : JavaStaticField
        Constant for making the layout extend under the navigation bar.
    SYSTEM_UI_FLAG_LAYOUT_STABLE : JavaStaticField
        Constant for preventing layout changes when the system UI
        visibility changes.
    SYSTEM_UI_FLAG_LIGHT_NAVIGATION_BAR : JavaStaticField
        Constant for enabling light navigation bar mode.
    SYSTEM_UI_FLAG_LIGHT_STATUS_BAR : JavaStaticField
        Constant for enabling light status bar mode.
    SYSTEM_UI_FLAG_LOW_PROFILE : JavaStaticField
        Constant for enabling low-profile mode for system UI.
    SYSTEM_UI_FLAG_VISIBLE : JavaStaticField
        Constant for ensuring the system UI is visible.
    SYSTEM_UI_LAYOUT_FLAGS : JavaStaticField
        Combined constant for specifying multiple layout-related flags.
    """

    __javaclass__ = 'android/view/View'
    SYSTEM_UI_FLAG_FULLSCREEN = JavaStaticField('I')
    SYSTEM_UI_FLAG_HIDE_NAVIGATION = JavaStaticField('I')
    SYSTEM_UI_FLAG_IMMERSIVE = JavaStaticField('I')
    SYSTEM_UI_FLAG_IMMERSIVE_STICKY = JavaStaticField('I')
    SYSTEM_UI_FLAG_LAYOUT_FULLSCREEN = JavaStaticField('I')
    SYSTEM_UI_FLAG_LAYOUT_HIDE_NAVIGATION = JavaStaticField('I')
    SYSTEM_UI_FLAG_LAYOUT_STABLE = JavaStaticField('I')
    SYSTEM_UI_FLAG_LIGHT_NAVIGATION_BAR = JavaStaticField('I')
    SYSTEM_UI_FLAG_LIGHT_STATUS_BAR = JavaStaticField('I')
    SYSTEM_UI_FLAG_LOW_PROFILE = JavaStaticField('I')
    SYSTEM_UI_FLAG_VISIBLE = JavaStaticField('I')
    SYSTEM_UI_LAYOUT_FLAGS = JavaStaticField('I')


class KeyEvent(JavaClass, metaclass=MetaJavaClass):
    """
    Represents a wrapper for the Android KeyEvent class.

    This class serves as a Python representation of the Android KeyEvent Java class, providing access
    to Java static fields, and facilitating interaction with Android key event constants and properties.

    Attributes:
    KEYCODE_BACK : int
        Static field representing the key code for the "Back" button in Android.
    """

    __javaclass__ = 'android/view/KeyEvent'
    KEYCODE_BACK = JavaStaticField('I')
    ACTION_DOWN = JavaStaticField('I')

    getAction = JavaMethod('()I')


class Choreographer(JavaClass, metaclass=MetaJavaClass):
    __javaclass__ = 'android/view/Choreographer'
    getInstance = JavaStaticMethod('()Landroid/view/Choreographer;')
    postFrameCallback = JavaMethod('(Landroid/view/Choreographer$FrameCallback;)V')
    removeFrameCallback = JavaMethod('(Landroid/view/Choreographer$FrameCallback;)V')



from copy import copy


class Request:
    def __init__(self, url: str, method: str, headers: dict) -> None:
        self.url = url
        self.method = method
        self.headers = copy(headers)

    def __repr__(self) -> str:
        return f'Request to {self.url} - {self.method}'


class Response:
    def __init__(self, url: str, status_code: int, headers: dict) -> None:
        self.url = url
        self.status_code = status_code
        self.headers = headers

    def __repr__(self) -> str:
        return f'Response from {self.url} - {self.status_code}'



import json
from enum import Enum
from typing import Any, Dict, Optional, Union

from webview.util import css_to_camel, escape_string


class DOMPropType(Enum):
    Style = 1
    Attribute = 2


class PropsDict:
    def __init__(self, element, type: DOMPropType, props: Optional[Dict[str, Any]] = None):
        self.__element = element
        self.__type = type

        if not props:
            return

        if type == DOMPropType.Style:
            converted_style = json.dumps({css_to_camel(key): value for key, value in props.items()})
            self.__element._window.evaluate_js(
                f"""
                {self.__element._query_command};
                var styles = JSON.parse('{escape_string(converted_style)}');

                for (var key in styles) {{
                    element.style[key] = styles[key];
                }}
            """
            )

        elif type == DOMPropType.Attribute:
            converted_attributes = json.dumps(
                {escape_string(key): escape_string(value) for key, value in props.items()}
            )

            self.__element._window.evaluate_js(
                f"""
                {self.__element._query_command};
                var attributes = JSON.parse('{converted_attributes}');

                for (var key in attributes) {{
                    if (key === 'data-pywebview-id') {{
                        continue;
                    }} else if (attributes[key] === null || attributes[key] === undefined) {{
                        element.removeAttribute(key);
                    }} else {{
                        element.setAttribute(key, attributes[key]);
                    }}
                }};
            """
            )

    def __getitem__(self, key):
        data = self.__get_data()
        return data.get(key)

    def __setitem__(self, key, value):
        if self.__type == DOMPropType.Style:
            self.__set_style({key: value})
        elif self.__type == DOMPropType.Attribute:
            self.__set_attribute({key: value})

    def __delitem__(self, key):
        if self.__type == DOMPropType.Style:
            self.__set_style({key: ''})
        elif self.__type == DOMPropType.Attribute:
            self.__set_attribute({key: None})

    def __contains__(self, key):
        data = self.__get_data()
        return key in data

    def keys(self):
        data = self.__get_data()
        return data.keys()

    def values(self):
        data = self.__get_data()
        return data.values()

    def items(self):
        data = self.__get_data()
        return data.items()

    def get(self, key, default=None):
        data = self.__get_data()
        return data.get(key, default)

    def clear(self):
        data = self.__get_data()

        for key in data.keys():
            data[key] = '' if self.__type == DOMPropType.Style else None

        if self.__type == DOMPropType.Style:
            self.__set_style(data)
        elif self.__type == DOMPropType.Attribute:
            self.__set_attribute(data)

    def copy(self):
        return self.__get_data()

    def update(self, other_dict: Dict[str, Union[str, int, float, None]]):
        if self.__type == DOMPropType.Style:
            self.__set_style(other_dict)
        elif self.__type == DOMPropType.Attribute:
            self.__set_attribute(other_dict)

    def pop(self, key, default=None):
        data = self.__get_data()
        return data.pop(key, default)

    def popitem(self):
        data = self.__get_data()
        return data.popitem()

    def __str__(self):
        data = self.__get_data()
        return str(data)

    def __repr__(self):
        data = self.__get_data()
        return repr(data)

    def __get_data(self) -> Dict[str, Any]:
        if self.__type == DOMPropType.Style:
            return self.__get_style()
        elif self.__type == DOMPropType.Attribute:
            return self.__get_attributes()

    def __get_attributes(self) -> Dict[str, Any]:
        return self.__element._window.evaluate_js(
            f"""
            {self.__element._query_command};
            var attributes = element.attributes;
            var result = {{}};
            for (var i = 0; i < attributes.length; i++) {{
                if (attributes[i].name === 'data-pywebview-id') {{
                    continue;
                }}
                result[attributes[i].name] = attributes[i].value;
            }}
            result
        """
        )

    def __set_attribute(self, props: Dict[str, Any]):
        self.__element._window.evaluate_js(
            f"""
            {self.__element._query_command};
            var values = JSON.parse('{escape_string(json.dumps(props))}');

            for (var key in values) {{
                var value = values[key];
                if (value === null || value === undefined) {{
                    element.removeAttribute(key);
                }} else {{
                    element.setAttribute(key, value);
                }}
            }}
        """
        )

    def __get_style(self) -> Dict[str, Any]:
        return self.__element._window.evaluate_js(
            f"""
            {self.__element._query_command};
            var styles = window.getComputedStyle(element);
            var computedStyles = {{}};

            for (var i = 0; i < styles.length; i++) {{
                var propertyName = styles[i];
                var propertyValue = styles.getPropertyValue(propertyName);

                if (propertyValue !== '') {{
                    computedStyles[propertyName] = propertyValue;
                }}
            }}

            computedStyles;
        """
        )

    def __set_style(self, style: Dict[str, Any]):
        converted_style = json.dumps({css_to_camel(key): value for key, value in style.items()})
        self.__element._window.evaluate_js(
            f"""
            {self.__element._query_command};
            var styles = JSON.parse('{escape_string(converted_style)}');

            for (var key in styles) {{
                var value = styles[key];
                element.style[key] = value;
            }}
        """
        )



from threading import Semaphore

from android.activity import register_activity_lifecycle_callbacks  # noqa
from android.runnable import run_on_ui_thread  # noqa

from webview.platforms.android.event import EventDispatcher
from webview.platforms.android.jclass.view import Choreographer
from webview.platforms.android.jinterface.view import FrameCallback


class EventLoop(EventDispatcher):
    def __init__(self):
        super().__init__()
        from webview.platforms.android.app import App

        self.app = App.get_running_app()
        self.quit = False
        self.status = 'idle'
        self.resumed = False
        self.destroyed = False
        self.paused = False
        register_activity_lifecycle_callbacks(
            onActivityCreated=self.app.on_create,
            onActivityPaused=self.app.on_pause,
            onActivityDestroyed=self.app.on_destroy,
            onActivityResumed=self.app.on_resume,
            onActivityStarted=self.app.on_start,
            onActivityStopped=self.app.on_stop,
        )

    def mainloop(self):
        choreographer = None
        frame_callback = None
        while not self.quit and self.status == 'created':

            def do_frame(_):
                lock.release()

            @run_on_ui_thread
            def post_frame():
                nonlocal choreographer, frame_callback

                if not choreographer:
                    choreographer = Choreographer.getInstance()
                if not frame_callback:
                    frame_callback = FrameCallback(do_frame)

                choreographer.postFrameCallback(frame_callback)

            lock = Semaphore(0)
            post_frame()
            lock.acquire()

    def close(self):
        self.quit = True
        self.status = 'destroyed'



__all__ = ('EventCallbackWrapper', 'JsApiCallbackWrapper', 'RequestInterceptor')

from jnius import PythonJavaClass, java_method


class EventCallbackWrapper(PythonJavaClass):
    """
    Represents a wrapper for event callbacks between Python and Java.

    This class enables interaction between Python and Java by implementing
    a Java interface and providing a callback mechanism. The purpose is to
    allow Python to respond to events triggered from the Java side by
    passing the event and related data.

    Attributes:
    callback (Callable[[str, str], None]): A callable function that handles
    an event and its associated data.

    Methods:
    callback(event: str, data: str): Calls the provided callback function
    with the given event and data.
    """

    __javacontext__ = 'app'
    __javainterfaces__ = ['com/pywebview/EventCallbackWrapper']

    def __init__(self, callback):
        self.callback_wrapper = callback

    @java_method('(Ljava/lang/String;Ljava/lang/String;)V')
    def callback(self, event, data):
        self.callback_wrapper(event, data)


class JsApiCallbackWrapper(PythonJavaClass):
    """
    Wrapper class for JavaScript API callbacks.

    This class serves as a bridge between a Python application and a Java
    environment, enabling communication through JavaScript API callback functionality.
    It implements the `JsApiCallbackWrapper` interface and is designed specifically
    for use with webview integration.

    Attributes:
    callback (Callable): A Python callable that gets executed when the callback
    method is triggered. The callable should accept three arguments.

    """

    __javacontext__ = 'app'
    __javainterfaces__ = ['com/pywebview/JsApiCallbackWrapper']

    def __init__(self, callback):
        self.callback_wrapper = callback

    @java_method('(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V')
    def callback(self, func, params, id):
        self.callback_wrapper(func, params, id)


class RequestInterceptor(PythonJavaClass):
    """
    Represents a class for intercepting web view requests and responses.

    This class provides a mechanism to handle events during web view
    requests and responses by allowing custom callback functions to be
    invoked. It acts as a bridge between Python and Java for facilitating
    interception of web-related actions within a web view environment.

    Attributes:
        on_request (Callable[[str, str, str], str]): Callback function that gets
            invoked when a web request is intercepted. The callback function should
            accept the request URL, HTTP method, and headers in JSON format, and it
            should return a string.
        on_response (Callable[[str, int, str], None]): Callback function that gets
            invoked when a web response is intercepted. The callback function should
            accept the response URL, HTTP status code, and headers in JSON format.

    """

    __javacontext__ = 'app'
    __javainterfaces__ = ['com/pywebview/WebViewRequestInterceptor']

    def __init__(self, on_request, on_response):
        self.on_request = on_request
        self.on_response = on_response

    @java_method('(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;')
    def onRequest(self, url: str, method: str, headers_json: str):
        return self.on_request(url, method, headers_json)

    @java_method('(Ljava/lang/String;ILjava/lang/String;)V')
    def onResponse(self, url: str, status_code: int, headers_json: str):
        self.on_response(url, status_code, headers_json)



import ctypes
import logging
import os
import signal
import sys
import tempfile
import threading
import winreg
from ctypes import windll, wintypes
from platform import machine
from threading import Event, Semaphore

try:
    import clr
except Exception:
    os.environ['PYTHONNET_RUNTIME'] = 'coreclr'
    import clr

from webview import FileDialog, _state, settings, windows
from webview.guilib import forced_gui_
from webview.menu import Menu, MenuAction, MenuSeparator
from webview.platforms import win32
from webview.screen import Screen
from webview.util import inject_base_uri, parse_file_type
from webview.window import FixPoint

clr.AddReference('System.Windows.Forms')
clr.AddReference('System.Collections')
clr.AddReference('System.Threading')
clr.AddReference('System.Reflection')

import System.Windows.Forms as WinForms  # noqa: E402
from Microsoft.Win32 import SystemEvents  # noqa: E402
from System import Array, Environment, Func, Int32, IntPtr, Object, Type, UInt32  # noqa: E402
from System.Drawing import Color, ColorTranslator, Icon, Point, Size, SizeF  # noqa: E402
from System.Reflection import Assembly, BindingFlags  # noqa: E402
from System.Threading import ApartmentState, Thread, ThreadStart  # noqa: E402

kernel32 = ctypes.WinDLL('kernel32', use_last_error=True)
logger = logging.getLogger('pywebview')
cache_dir = None
_sigint_received = False


def _sigint_handler(signum, frame):
    """
    Handler for SIGINT signal (Ctrl+C).

    Sets a flag that's checked by the timer in the GUI thread. This is necessary because
    the signal handler runs in the main thread, but Application.Exit() must be called
    from the GUI thread. The timer in create_window() checks this flag periodically and
    exits the application when it's set.
    """
    global _sigint_received
    _sigint_received = True


def _is_new_version(current_version: str, new_version: str) -> bool:
    new_range = new_version.split('.')
    cur_range = current_version.split('.')
    for index, _ in enumerate(new_range):
        if len(cur_range) > index:
            return int(new_range[index]) >= int(cur_range[index])

    return False


def _is_chromium():
    if settings['WEBVIEW2_RUNTIME_PATH']:
        return True

    def edge_build(key_type, key, description=''):
        try:
            if machine() == 'x86' or key_type == 'HKEY_CURRENT_USER':
                path = rf'Microsoft\EdgeUpdate\Clients\{key}'
            else:
                path = rf'WOW6432Node\Microsoft\EdgeUpdate\Clients\{key}'

            with winreg.OpenKey(getattr(winreg, key_type), rf'SOFTWARE\{path}') as windows_key:
                build, _ = winreg.QueryValueEx(windows_key, 'pv')
                return str(build)

        except Exception:
            pass

        return '0'

    try:
        net_key = winreg.OpenKey(
            winreg.HKEY_LOCAL_MACHINE, r'SOFTWARE\Microsoft\NET Framework Setup\NDP\v4\Full'
        )
        version, _ = winreg.QueryValueEx(net_key, 'Release')

        if version < 394802:  # .NET 4.6.2
            return False

        build_versions = [
            {
                'key': '{F3017226-FE2A-4295-8BDF-00C3A9A7E4C5}',
                'description': 'Microsoft Edge WebView2 Runtime',
            },  # runtime
            {
                'key': '{2CD8A007-E189-409D-A2C8-9AF4EF3C72AA}',
                'description': 'Microsoft Edge WebView2 Beta',
            },  # beta
            {
                'key': '{0D50BFEC-CD6A-4F9A-964C-C7416E3ACB10}',
                'description': 'Microsoft Edge WebView2 Developer',
            },  # dev
            {
                'key': '{65C35B14-6C1D-4122-AC46-7148CC9D6497}',
                'description': 'Microsoft Edge WebView2 Canary',
            },  # canary
        ]

        for item in build_versions:
            for key_type in ('HKEY_CURRENT_USER', 'HKEY_LOCAL_MACHINE'):
                build = edge_build(key_type, item['key'], item['description'])
                if _is_new_version('86.0.622.0', build):  # Webview2 86.0.622.0
                    return True

    except Exception as e:
        logger.exception(e)
    finally:
        winreg.CloseKey(net_key)

    return False


is_cef = forced_gui_ == 'cef'
is_chromium = not is_cef and _is_chromium() and forced_gui_ != 'mshtml'

if is_cef:
    from . import cef as CEF

    IWebBrowserInterop = object

    logger.debug('Using WinForms / CEF')
    renderer = 'cef'
elif is_chromium:
    from . import edgechromium as Chromium

    IWebBrowserInterop = object

    logger.debug('Using WinForms / Chromium')
    renderer = 'edgechromium'
else:
    from . import mshtml as IE

    logger.warning(
        'MSHTML is deprecated. See https://pywebview.flowrl.com/guide/web_engine.html on details how to use Edge Chromium'
    )
    logger.debug('Using WinForms / MSHTML')
    IE._set_ie_mode()
    renderer = 'mshtml'


def DwmSetWindowAttribute(hwnd, attr, value, size=4):
    DwmSetWindowAttribute = ctypes.windll.dwmapi.DwmSetWindowAttribute
    DwmSetWindowAttribute.argtypes = [
        wintypes.HWND,
        wintypes.DWORD,
        ctypes.c_void_p,
        wintypes.DWORD,
    ]
    return DwmSetWindowAttribute(hwnd, attr, ctypes.byref(ctypes.c_int(value)), size)


def ExtendFrameIntoClientArea(hwnd):
    class _MARGINS(ctypes.Structure):
        _fields_ = [
            ('cxLeftWidth', ctypes.c_int),
            ('cxRightWidth', ctypes.c_int),
            ('cyTopHeight', ctypes.c_int),
            ('cyBottomHeight', ctypes.c_int),
        ]

    DwmExtendFrameIntoClientArea = ctypes.windll.dwmapi.DwmExtendFrameIntoClientArea
    m = _MARGINS()
    m.cxLeftWidth = 1
    m.cxRightWidth = 1
    m.cyTopHeight = 1
    m.cyBottomHeight = 1
    return DwmExtendFrameIntoClientArea(hwnd, ctypes.byref(m))


class BrowserView:
    instances = {}

    class BrowserForm(WinForms.Form):
        def __init__(self, window, cache_dir):
            super().__init__()
            self.uid = window.uid
            self.pywebview_window = window
            self.pywebview_window.native = self
            self.real_url = None
            self.Text = window.title
            # Store initial window size as logical pixels, will be converted later
            self._initial_width = window.initial_width
            self._initial_height = window.initial_height

            self.AutoScaleDimensions = SizeF(96.0, 96.0)
            self.AutoScaleMode = WinForms.AutoScaleMode.Dpi
            hwnd = self.Handle.ToInt32()

            # Set the initial size now that we have a window handle and can get DPI
            # Size and MinimumSize need to be in physical pixels
            scale = self._scale
            self.Size = Size(int(window.initial_width * scale), int(window.initial_height * scale))
            self.MinimumSize = Size(
                int(window.min_size[0] * scale), int(window.min_size[1] * scale)
            )

            if window.initial_x is not None and window.initial_y is not None:
                self.StartPosition = WinForms.FormStartPosition.Manual
                # Convert logical pixel coordinates to physical for WinForms
                self.Location = Point(
                    int(window.initial_x * scale),
                    int(window.initial_y * scale),
                )
            elif window.screen:
                self.StartPosition = WinForms.FormStartPosition.Manual
                # Screen coordinates are in logical pixels, center the window
                # Calculate center position in logical pixels first, then convert to physical
                logical_x = window.screen.x + (window.screen.width - window.initial_width) // 2
                logical_y = window.screen.y + (window.screen.height - window.initial_height) // 2
                self.Location = Point(int(logical_x * scale), int(logical_y * scale))
            else:
                self.StartPosition = WinForms.FormStartPosition.CenterScreen

            if not window.resizable:
                self.FormBorderStyle = WinForms.FormBorderStyle.FixedSingle
                self.MaximizeBox = False

            if window.maximized:
                self.WindowState = WinForms.FormWindowState.Maximized
            elif window.minimized:
                self.WindowState = WinForms.FormWindowState.Minimized

            self.old_state = self.WindowState

            # Application icon
            if _state['icon'] and os.path.isfile(_state['icon']):
                self.Icon = Icon(_state['icon'])
            else:
                handle = kernel32.GetModuleHandleW(None)
                icon_handle = windll.shell32.ExtractIconW(handle, sys.executable, 0)

                if icon_handle != 0:
                    self.Icon = Icon.FromHandle(IntPtr.op_Explicit(Int32(icon_handle))).Clone()
                    windll.user32.DestroyIcon(icon_handle)

            self.closed = window.events.closed
            self.closing = window.events.closing
            self.shown = window.events.shown
            self.loaded = window.events.loaded
            self.url = window.real_url
            self.TopMost = window.on_top

            self.is_fullscreen = False
            if window.fullscreen:
                self.toggle_fullscreen()

            if window.shadow and not window.transparent:
                # Should do this before set frameless
                ExtendFrameIntoClientArea(hwnd)
                DwmSetWindowAttribute(hwnd, 2, 2, 4)

            if window.frameless:
                self.frameless = window.frameless
                self.FormBorderStyle = getattr(WinForms.FormBorderStyle, 'None')

            if window.menu or _state['menu']:
                self.set_window_menu(window.menu or _state['menu'])

            if is_cef:
                self.browser = None
                CEF.create_browser(window, hwnd, BrowserView.alert, self)
            elif is_chromium:
                self.browser = Chromium.EdgeChrome(self, window, cache_dir)
                self.webview = self.browser.webview
            else:
                self.browser = IE.MSHTML(self, window, BrowserView.alert)
                self.webview = self.browser.webview

            if (
                window.transparent and self.browser
            ):  # window transparency is supported only with EdgeChromium
                self.SetStyle(WinForms.ControlStyles.SupportsTransparentBackColor, True)
                self.browser.DefaultBackgroundColor = Color.Transparent
            else:
                self.BackColor = ColorTranslator.FromHtml(window.background_color)

            if not window.focus:
                windll.user32.SetWindowLongW(
                    self.Handle.ToInt32(),
                    -20,
                    windll.user32.GetWindowLongW(self.Handle.ToInt32(), -20) | 0x8000000,
                )

            self.Activated += self.on_activated
            self.Shown += self.on_shown
            self.FormClosed += self.on_close
            self.FormClosing += self.on_closing
            self.Resize += self.on_resize
            self.Move += self.on_move

            self.localization = window.localization

            self.update_title_bar_theme()
            SystemEvents.UserPreferenceChanged += self.on_system_theme_changed

        def __str__(self):
            return f'<System.Windows.Forms object with {self.Handle} handle>'

        @property
        def _scale(self):
            """Logical-to-physical pixel scale for the monitor this window is on."""
            if is_chromium:
                try:
                    # Use per-window DPI for accurate multi-monitor support
                    return windll.user32.GetDpiForWindow(self.Handle.ToInt32()) / 96
                except Exception as e:
                    logger.warning(f'Failed to get DPI for window: {e}')
                    return 1.0
            else:
                # MSHTML doesn't need scaling
                return 1.0

        def on_system_theme_changed(self, sender, e):
            self.update_title_bar_theme()

        def update_title_bar_theme(self):
            if self.is_dark_theme():
                DwmSetWindowAttribute(self.Handle.ToInt32(), 20, 1)
                DwmSetWindowAttribute(self.Handle.ToInt32(), 38, 2)
            else:
                DwmSetWindowAttribute(self.Handle.ToInt32(), 20, 0)
                DwmSetWindowAttribute(self.Handle.ToInt32(), 38, 1)

        def is_dark_theme(self):
            try:
                personalize_key = winreg.OpenKey(
                    winreg.HKEY_CURRENT_USER,
                    r'Software\Microsoft\Windows\CurrentVersion\Themes\Personalize',
                    0,
                    winreg.KEY_READ,
                )
                system_theme, _ = winreg.QueryValueEx(personalize_key, 'AppsUseLightTheme')
                winreg.CloseKey(personalize_key)
                if system_theme == 0:
                    return True
                else:
                    return False
            except Exception as e:
                logger.debug(f'Error while getting system theme: {e}')
                return None

        def on_activated(self, *_):
            if not self.pywebview_window.focus:
                windll.user32.SetWindowLongW(
                    self.Handle.ToInt32(),
                    -20,
                    windll.user32.GetWindowLongW(self.Handle.ToInt32(), -20) | 0x8000000,
                )

        def on_shown(self, *_):
            if not is_cef:
                self.shown.set()
                self.browser.webview.Focus()

        def on_close(self, *_):
            def _shutdown():
                if is_cef:
                    CEF.shutdown()
                elif is_chromium:
                    self.hide()
                    self.browser.clear_user_data()

                WinForms.Application.Exit()

            if not is_cef:
                # stop waiting for JS result
                self.browser.js_result_semaphore.release()

            if is_cef:
                CEF.close_window(self.uid)

            del BrowserView.instances[self.uid]

            # during tests windows is empty for some reason. no idea why.
            if self.pywebview_window in windows:
                windows.remove(self.pywebview_window)

            self.closed.set()

            if len(BrowserView.instances) == 0:
                self.Invoke(Func[Type](_shutdown))

        def on_closing(self, sender, args):
            should_cancel = self.closing.set()
            if should_cancel:
                args.Cancel = True

            if not args.Cancel:
                if self.pywebview_window.confirm_close:
                    result = WinForms.MessageBox.Show(
                        self.localization['global.quitConfirmation'],
                        self.Text,
                        WinForms.MessageBoxButtons.OKCancel,
                        WinForms.MessageBoxIcon.Asterisk,
                    )

                    if result == WinForms.DialogResult.Cancel:
                        args.Cancel = True

        def on_resize(self, sender, args):
            if self.WindowState == WinForms.FormWindowState.Maximized:
                self.pywebview_window.events.maximized.set()

            if self.WindowState == WinForms.FormWindowState.Minimized:
                self.pywebview_window.events.minimized.set()

            if self.WindowState == WinForms.FormWindowState.Normal and self.old_state in (
                WinForms.FormWindowState.Minimized,
                WinForms.FormWindowState.Maximized,
            ):
                self.pywebview_window.events.restored.set()

            self.old_state = self.WindowState

            if is_cef:
                CEF.resize(self.Width, self.Height, self.uid)

            # Convert physical pixel dimensions to logical pixels for the API
            scale = self._scale
            self.pywebview_window.events.resized.set(
                int(self.Width / scale), int(self.Height / scale)
            )

        def on_move(self, sender, args):
            # Convert physical pixel location to logical pixels for the API
            scale = self._scale
            self.pywebview_window.events.moved.set(
                int(self.Location.X / scale),
                int(self.Location.Y / scale),
            )

        def evaluate_js(self, script, parse_json):
            result = self.browser.evaluate_js(script, parse_json)
            return result

        def clear_cookies(self):
            def _clear_cookies():
                self.browser.clear_cookies()

            if not is_chromium:
                logger.error('clear_cookies() is not implemented for this platform')
                return

            self.Invoke(Func[Type](_clear_cookies))

        def get_cookies(self):
            def _get_cookies():
                self.browser.get_cookies(cookies, semaphore)

            cookies = []
            if not is_chromium:
                logger.error('get_cookies() is not implemented for this platform')
                return cookies

            semaphore = Semaphore(0)

            self.Invoke(Func[Type](_get_cookies))
            semaphore.acquire()

            return cookies

        def load_html(self, content, base_uri):
            def _load_html():
                self.browser.load_html(content, base_uri)

            self.Invoke(Func[Type](_load_html))

        def load_url(self, url):
            def _load_url():
                self.browser.load_url(url)

            self.Invoke(Func[Type](_load_url))

        def hide(self):
            self.Invoke(Func[Type](self.Hide))

        def show(self):
            def _show():
                self.Show()
                self.Activate()

            if self.InvokeRequired:
                self.Invoke(Func[Type](_show))
            else:
                _show()

        def set_window_menu(self, menu_list):
            def _set_window_menu():
                def create_action_item(menu_line_item):
                    action_item = WinForms.ToolStripMenuItem(menu_line_item.title)
                    # Don't run action function on main thread
                    action_item.Click += (
                        lambda _, __, menu_line_item=menu_line_item: threading.Thread(
                            target=menu_line_item.function
                        ).start()
                    )
                    return action_item

                def create_submenu(title, line_items, supermenu=None):
                    m = WinForms.ToolStripMenuItem(title)
                    for menu_line_item in line_items:
                        if isinstance(menu_line_item, MenuSeparator):
                            m.DropDownItems.Add(WinForms.ToolStripSeparator())
                            continue
                        elif isinstance(menu_line_item, MenuAction):
                            m.DropDownItems.Add(create_action_item(menu_line_item))
                        elif isinstance(menu_line_item, Menu):
                            create_submenu(menu_line_item.title, menu_line_item.items, m)

                    if supermenu:
                        supermenu.DropDownItems.Add(m)

                    return m

                top_level_menu = WinForms.MenuStrip()

                for menu in menu_list:
                    # Ignore '__app__' menus (macOS-only feature)
                    if isinstance(menu, Menu) and menu.title == '__app__':
                        continue
                    if isinstance(menu, Menu):
                        top_level_menu.Items.Add(create_submenu(menu.title, menu.items))
                    elif isinstance(menu, MenuAction):
                        top_level_menu.Items.Add(create_action_item(menu))

                self.Controls.Add(top_level_menu)

            if self.InvokeRequired:
                self.Invoke(Func[Type](_set_window_menu))
            else:
                _set_window_menu()

        def toggle_fullscreen(self):
            def _toggle():
                if not self.is_fullscreen:
                    self.old_size = self.Size
                    self.old_state = self.WindowState
                    self.old_style = self.FormBorderStyle
                    self.old_location = self.Location
                    self.old_screen = WinForms.Screen.FromControl(self)
                    self.FormBorderStyle = getattr(WinForms.FormBorderStyle, 'None')
                    self.Bounds = WinForms.Screen.FromControl(self).Bounds
                    self.WindowState = WinForms.FormWindowState.Maximized
                    self.is_fullscreen = True
                    windll.user32.SetWindowPos(
                        self.Handle.ToInt32(),
                        None,
                        self.old_screen.Bounds.X,
                        self.old_screen.Bounds.Y,
                        self.old_screen.Bounds.Width,
                        self.old_screen.Bounds.Height,
                        64,
                    )
                    # disable window rounding
                    DwmSetWindowAttribute(self.Handle.ToInt32(), 33, 1)
                    # hide window border
                    DwmSetWindowAttribute(self.Handle.ToInt32(), 34, 0xFFFFFFFE)
                else:
                    self.WindowState = WinForms.FormWindowState.Normal
                    self.FormBorderStyle = self.old_style
                    self.is_fullscreen = False
                    valid_location = any(
                        screen == self.old_screen for screen in WinForms.Screen.AllScreens
                    )

                    if not valid_location:
                        self.Size = self.old_size
                        self.CenterToScreen()
                    else:
                        self.Location = self.old_location
                        self.Size = self.old_size

                    # enable window rounding
                    DwmSetWindowAttribute(self.Handle.ToInt32(), 33, 0)
                    # show window border
                    DwmSetWindowAttribute(self.Handle.ToInt32(), 34, 0xFFFFFFFF)

            if self.InvokeRequired:
                self.Invoke(Func[Type](_toggle))
            else:
                _toggle()

        def resize(self, width, height, fix_point):
            # Input width/height are in logical pixels, need to convert to physical
            scale = self._scale
            phys_width = int(width * scale)
            phys_height = int(height * scale)

            # Location is already in physical pixels
            x = self.Location.X
            y = self.Location.Y

            if fix_point & FixPoint.EAST:
                x = x + self.Width - phys_width

            if fix_point & FixPoint.SOUTH:
                y = y + self.Height - phys_height

            windll.user32.SetWindowPos(
                self.Handle.ToInt32(), None, x, y, phys_width, phys_height, 64
            )

        def move(self, x, y):
            # Input x/y are in logical pixels, need to convert to physical
            SWP_NOSIZE = 0x0001  # Retains the current size
            SWP_NOZORDER = 0x0004  # Retains the current Z order
            SWP_SHOWWINDOW = 0x0040  # Displays the window

            scale = self._scale
            if scale != 1:
                # Convert logical pixels to physical pixels
                x_phys = int(x * scale)
                y_phys = int(y * scale)
            else:
                x_phys = int(x)
                y_phys = int(y)

            windll.user32.SetWindowPos(
                self.Handle.ToInt32(),
                None,
                x_phys,
                y_phys,
                None,
                None,
                SWP_NOSIZE | SWP_NOZORDER | SWP_SHOWWINDOW,
            )

        def maximize(self):
            def _maximize():
                self.WindowState = WinForms.FormWindowState.Maximized

            self.Invoke(Func[Type](_maximize))

        def minimize(self):
            def _minimize():
                self.WindowState = WinForms.FormWindowState.Minimized

            self.Invoke(Func[Type](_minimize))

        def restore(self):
            def _restore():
                self.WindowState = WinForms.FormWindowState.Normal

            self.Invoke(Func[Type](_restore))

    @staticmethod
    def alert(message):
        WinForms.MessageBox.Show(str(message))


class OpenFolderDialog:
    foldersFilter = 'Folders|\n'
    flags = BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic
    windowsFormsAssembly = Assembly.LoadWithPartialName('System.Windows.Forms')
    iFileDialogType = windowsFormsAssembly.GetType(
        'System.Windows.Forms.FileDialogNative+IFileDialog'
    )
    OpenFileDialogType = windowsFormsAssembly.GetType('System.Windows.Forms.OpenFileDialog')
    FileDialogType = windowsFormsAssembly.GetType('System.Windows.Forms.FileDialog')
    createVistaDialogMethodInfo = OpenFileDialogType.GetMethod('CreateVistaDialog', flags)
    onBeforeVistaDialogMethodInfo = OpenFileDialogType.GetMethod('OnBeforeVistaDialog', flags)
    getOptionsMethodInfo = FileDialogType.GetMethod('GetOptions', flags)
    setOptionsMethodInfo = iFileDialogType.GetMethod('SetOptions', flags)
    fosPickFoldersBitFlag = (
        windowsFormsAssembly.GetType('System.Windows.Forms.FileDialogNative+FOS')
        .GetField('FOS_PICKFOLDERS')
        .GetValue(None)
    )

    vistaDialogEventsConstructorInfo = windowsFormsAssembly.GetType(
        'System.Windows.Forms.FileDialog+VistaDialogEvents'
    ).GetConstructor(flags, None, [FileDialogType], [])
    adviseMethodInfo = iFileDialogType.GetMethod('Advise')
    unadviseMethodInfo = iFileDialogType.GetMethod('Unadvise')
    showMethodInfo = iFileDialogType.GetMethod('Show')

    @classmethod
    def show(cls, parent=None, initialDirectory=None, allow_multiple=False, title=None):
        openFileDialog = WinForms.OpenFileDialog()
        openFileDialog.InitialDirectory = initialDirectory
        openFileDialog.Title = title
        openFileDialog.Filter = OpenFolderDialog.foldersFilter
        openFileDialog.AddExtension = False
        openFileDialog.CheckFileExists = False
        openFileDialog.DereferenceLinks = True
        openFileDialog.Multiselect = allow_multiple
        openFileDialog.RestoreDirectory = True

        iFileDialog = OpenFolderDialog.createVistaDialogMethodInfo.Invoke(openFileDialog, [])
        OpenFolderDialog.onBeforeVistaDialogMethodInfo.Invoke(openFileDialog, [iFileDialog])
        options = OpenFolderDialog.getOptionsMethodInfo.Invoke(openFileDialog, [])
        options = options.op_BitwiseOr(OpenFolderDialog.fosPickFoldersBitFlag)
        OpenFolderDialog.setOptionsMethodInfo.Invoke(iFileDialog, [options])
        adviseParametersWithOutputConnectionToken = Array[Object](
            [
                OpenFolderDialog.vistaDialogEventsConstructorInfo.Invoke([openFileDialog]),
                UInt32(0),
            ]
        )
        OpenFolderDialog.adviseMethodInfo.Invoke(
            iFileDialog, adviseParametersWithOutputConnectionToken
        )
        dwCookie = adviseParametersWithOutputConnectionToken.GetValue(1)
        try:
            result = OpenFolderDialog.showMethodInfo.Invoke(
                iFileDialog, [parent.Handle if parent else None]
            )
            if result == 0:
                return tuple(openFileDialog.FileNames)

            return None

        finally:
            OpenFolderDialog.unadviseMethodInfo.Invoke(iFileDialog, [UInt32(dwCookie)])


_main_window_created = Event()
_main_window_created.clear()

_already_set_up_app = False


def init_storage():
    global cache_dir

    if not _state['private_mode'] or _state['storage_path']:
        try:
            data_folder = Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData)

            if not os.access(data_folder, os.W_OK):
                data_folder = Environment.GetFolderPath(Environment.SpecialFolder.UserProfile)

            cache_dir = _state['storage_path'] or os.path.join(data_folder, 'pywebview')

            if not os.path.exists(cache_dir):
                os.makedirs(cache_dir)
        except Exception:
            logger.exception(f'Cache directory {cache_dir} creation failed')
    else:
        cache_dir = tempfile.TemporaryDirectory().name


def setup_app():
    # MUST be called before create_window
    global _already_set_up_app
    if _already_set_up_app:
        return

    WinForms.Application.EnableVisualStyles()
    WinForms.Application.SetCompatibleTextRenderingDefault(False)
    _already_set_up_app = True


def create_window(window):
    def create():
        browser = BrowserView.BrowserForm(window, cache_dir)
        BrowserView.instances[window.uid] = browser
        window.events.before_show.set()

        if window.hidden:
            browser.Opacity = 0
            browser.Show()
            browser.Hide()
            browser.Opacity = 1
        elif window.transparent and is_chromium:
            # hack to make transparent window work
            # window is started hidden and shown on Navigating event.
            # no idea why this works
            browser.Show()
            browser.Hide()
        else:
            browser.Show()

        _main_window_created.set()

        if window.uid == 'master':

            def timer_tick(sender, e):
                # Check if SIGINT was received and exit from GUI thread
                global _sigint_received
                if _sigint_received:
                    app.Exit()

            # Create a timer to periodically allow the Python interpreter to run
            # This enables the signal handler to be called even when the WinForms event loop is running
            timer = WinForms.Timer()
            timer.Interval = 500  # 500ms
            timer.Tick += timer_tick
            timer.Start()

            app.Run()

    app = WinForms.Application

    if window.uid == 'master':
        # Set up Ctrl+C handler in main thread (before starting GUI thread)
        signal.signal(signal.SIGINT, _sigint_handler)

        if is_chromium:
            init_storage()

        if sys.getwindowsversion().major >= 6:
            windll.user32.SetProcessDPIAware()

        if is_cef:
            CEF.init(window, cache_dir)

        thread = Thread(ThreadStart(create))
        thread.SetApartmentState(ApartmentState.STA)
        thread.Start()

        # Don't use thread.Join() as it blocks the main thread indefinitely,
        # preventing signal handlers from being processed. Instead, periodically
        # check if the thread is still alive, which allows the Python interpreter
        # to process pending signals (like SIGINT from Ctrl+C)
        while thread.IsAlive:
            thread.Join(500)

    else:
        _main_window_created.wait()
        i = list(BrowserView.instances.values())[0]  # arbitrary instance
        i.Invoke(Func[Type](create))


def set_title(title, uid):
    def _set_title():
        i.Text = title

    i = BrowserView.instances.get(uid)

    if not i:
        return
    elif i.InvokeRequired:
        i.Invoke(Func[Type](_set_title))
    else:
        _set_title()


def create_confirmation_dialog(title, message, uid):
    i = BrowserView.instances.get(uid)

    if not i:
        return

    result = WinForms.MessageBox.Show(message, title, WinForms.MessageBoxButtons.OKCancel)
    return result == WinForms.DialogResult.OK


def create_file_dialog(dialog_type, directory, allow_multiple, save_filename, file_types, uid):
    i = BrowserView.instances.get(uid)

    if not i:
        return

    if not directory:
        directory = os.environ['HOMEPATH']

    try:
        if dialog_type == FileDialog.FOLDER:
            file_path = OpenFolderDialog.show(i, directory, allow_multiple)

        elif dialog_type == FileDialog.OPEN:
            dialog = WinForms.OpenFileDialog()

            dialog.Multiselect = allow_multiple
            dialog.InitialDirectory = directory

            if len(file_types) > 0:
                dialog.Filter = '|'.join(
                    ['{0} ({1})|{1}'.format(*parse_file_type(f)) for f in file_types]
                )
            else:
                dialog.Filter = i.localization['windows.fileFilter.allFiles'] + ' (*.*)|*.*'
            dialog.RestoreDirectory = True

            result = dialog.ShowDialog(i)
            if result == WinForms.DialogResult.OK:
                file_path = tuple(dialog.FileNames)
            else:
                file_path = None

        elif dialog_type == FileDialog.SAVE:
            dialog = WinForms.SaveFileDialog()
            if len(file_types) > 0:
                dialog.Filter = '|'.join(
                    ['{0} ({1})|{1}'.format(*parse_file_type(f)) for f in file_types]
                )
            else:
                dialog.Filter = i.localization['windows.fileFilter.allFiles'] + ' (*.*)|*.*'
            dialog.InitialDirectory = directory
            dialog.RestoreDirectory = True
            dialog.FileName = save_filename

            result = dialog.ShowDialog(i)
            if result == WinForms.DialogResult.OK:
                file_path = (dialog.FileName,)
            else:
                file_path = None

        return file_path
    except Exception as e:
        logger.exception(f'Error invoking {dialog_type} dialog: {e}')
        return None


def clear_cookies(uid):
    if is_cef:
        CEF.clear_cookies(uid)
    i = BrowserView.instances.get(uid)

    if i:
        i.clear_cookies()


def get_cookies(uid):
    if is_cef:
        return CEF.get_cookies(uid)
    i = BrowserView.instances.get(uid)

    if i:
        return i.get_cookies()


def get_current_url(uid):
    if is_cef:
        return CEF.get_current_url(uid)

    i = BrowserView.instances.get(uid)
    if i:
        return i.browser.url


def load_url(url, uid):
    i = BrowserView.instances.get(uid)
    if not i:
        return

    if is_cef:
        CEF.load_url(url, uid)
    else:
        i.load_url(url)


def load_html(content, base_uri, uid):
    i = BrowserView.instances.get(uid)

    if is_cef:
        CEF.load_html(inject_base_uri(content, base_uri), uid)
    elif i:
        i.load_html(content, base_uri)


def get_active_window():
    active_window = None
    try:
        active_window = WinForms.Form.ActiveForm
    except Exception:
        return None

    if active_window:
        for uid, browser_view_instance in BrowserView.instances.items():
            if browser_view_instance.Handle == active_window.Handle:
                return browser_view_instance.pywebview_window

    return None


def show(uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.show()


def hide(uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.hide()


def toggle_fullscreen(uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.toggle_fullscreen()


def set_on_top(uid, on_top):
    i = BrowserView.instances.get(uid)
    if i:
        i.TopMost = on_top


def resize(width, height, uid, fix_point):
    i = BrowserView.instances.get(uid)
    if i:
        i.resize(width, height, fix_point)


def move(x, y, uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.move(x, y)


def maximize(uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.maximize()


def minimize(uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.minimize()


def restore(uid):
    i = BrowserView.instances.get(uid)
    if i:
        i.restore()


def destroy_window(uid):
    def _close():
        i.Close()

    i = BrowserView.instances.get(uid)
    if not i:
        return

    i.Invoke(Func[Type](_close))
    if not is_cef:
        i.browser.js_result_semaphore.release()


def evaluate_js(script, uid, parse_json, result_id=None):
    if is_cef:
        return CEF.evaluate_js(script, result_id, parse_json, uid)

    i = BrowserView.instances.get(uid)
    if i:
        return i.evaluate_js(script, parse_json)


def get_position(uid):
    i = BrowserView.instances.get(uid)
    if i:
        # Convert physical pixel position to logical pixels
        scale = i._scale
        return int(i.Left / scale), int(i.Top / scale)


def get_size(uid):
    i = BrowserView.instances.get(uid)
    if i:
        # Size is in physical pixels, convert to logical pixels
        size = i.Size
        scale = i._scale
        return int(size.Width / scale), int(size.Height / scale)


def get_screens():
    """Get all screens with coordinates in logical pixels."""
    screens = []

    for s in WinForms.Screen.AllScreens:
        # Get logical size from Bounds (already DPI-aware scaled due to SetProcessDPIAware)
        logical_width = s.Bounds.Width
        logical_height = s.Bounds.Height

        # Get scale factor by comparing physical vs logical resolution
        scale = win32.get_screen_scale(s.DeviceName, logical_width, logical_height)

        # Bounds are already in logical pixels due to SetProcessDPIAware
        screens.append(
            Screen(
                s.Bounds.X,
                s.Bounds.Y,
                logical_width,
                logical_height,
                s.WorkingArea,
                scale,
            )
        )

    return screens


def add_tls_cert(_):
    return



__all__ = ('Context',)

from jnius import JavaClass, JavaStaticField, MetaJavaClass


class Context(JavaClass, metaclass=MetaJavaClass):
    """
    Represents the Android Context Java class providing access to application-specific
    resources and services.

    This class serves as a bridge to interact with the Android framework and various
    system services provided by the Android operating system. It holds functionality
    for accessing application environments, managing resources, and invoking system:
    defined services.

    Attributes:
        __javaclass__ (str): The fully qualified Java class name representation of the Context class.
        DOWNLOAD_SERVICE (JavaStaticField): A static field representing the Java constant
            for the download service.
    """

    __javaclass__ = 'android/content/Context'
    DOWNLOAD_SERVICE = JavaStaticField('Ljava/lang/String;')



__all__ = ('ValueCallback', 'DownloadListener')

from jnius import PythonJavaClass, java_method


class ValueCallback(PythonJavaClass):
    """
    Represents a callback to receive a value from Java code.

    This class implements the Android ValueCallback interface in order to receive
    values from a Java environment. It's particularly useful when working with
    Android-related tasks where results from Java methods are passed back to Python.

    Attributes:
    on_receive_value: Callable function to handle the received value from the Java
    environment.

    Methods:
    onReceiveValue: This method is invoked when a value is received from Java,
    passing the value to the Python callback function.
    """

    __javacontext__ = 'app'
    __javainterfaces__ = ['android/webkit/ValueCallback']

    def __init__(self, on_receive_value):
        self.on_receive_value = on_receive_value

    @java_method('(Ljava/lang/Object;)V')
    def onReceiveValue(self, value):
        """
        Handles the receiving of a single value. This method is invoked to process or
        handle a given object value when required.

        Args:
            value: The object received which needs to be processed.
        """
        self.on_receive_value(value)


class DownloadListener(PythonJavaClass):
    """
    Handles download events initiated through a WebView in an Android application.

    This class serves as a bridge for the Android `DownloadListener` interface,
    facilitating the handling of download requests generated during WebView interactions.
    By implementing this interface, it allows the developer to intercept and handle
    download events directly using the provided callback.

    Attributes:
    on_download_start (callable): A callback function that is triggered when a download
        starts. The function should accept five parameters: url, user_agent, content_disposition,
        mimetype, and content_length.
    """

    __javacontext__ = 'app'
    __javainterfaces__ = ['android/webkit/DownloadListener']

    def __init__(self, on_download_start):
        self.on_download_start = on_download_start

    @java_method('(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;J)V')
    def onDownloadStart(self, url, user_agent, content_disposition, mimetype, content_length):
        self.on_download_start(url, user_agent, content_disposition, mimetype, content_length)



__all__ = ('Uri',)

from jnius import JavaClass, JavaMethod, JavaStaticMethod, MetaJavaClass


class Uri(JavaClass, metaclass=MetaJavaClass):
    """
    Represents a URI reference as defined in RFC 2396, composed of components such as
    scheme, authority, path, query, and fragment.

    The Uri class provides methods to work with uniform resource identifiers (URI). It
    facilitates parsing, retrieving specific parts of the URI, and performing a variety
    of URI-related manipulations. Typically used in Android development for handling
    URI-based resources.

    Attributes:
        __javaclass__ : str
            Specifies the associated Java class for this Python wrapper.
        parse : JavaMethod
            Represents the Java method 'parse' to convert a string to a Uri instance.
        getLastPathSegment : JavaMethod
            Represents the Java method 'getLastPathSegment' to retrieve the last
            segment of the URI's path.
    """

    __javaclass__ = 'android/net/Uri'
    parse = JavaStaticMethod('(Ljava/lang/String;)Landroid/net/Uri;')
    getLastPathSegment = JavaMethod('()Ljava/lang/String;')



import json
import logging
import os
import shutil
import webbrowser
import winreg
from threading import Semaphore

try:
    import clr
except Exception:
    os.environ['PYTHONNET_RUNTIME'] = 'coreclr'
    import clr

from webview import Window, _state
from webview import settings as webview_settings
from webview.dom import _dnd_state
from webview.models import Request, Response
from webview.util import (
    DEFAULT_HTML,
    create_cookie,
    get_app_root,
    inject_pywebview,
    interop_dll_path,
    js_bridge_call,
)

clr.AddReference('System.Windows.Forms')
clr.AddReference('System.Collections')
clr.AddReference('System.Threading')

import System.Windows.Forms as WinForms  # noqa: E402
from System import Action, Convert, Func, Object, String, Type, Uri  # noqa: E402
from System.Collections.Generic import List  # noqa: E402
from System.Diagnostics import Process  # noqa: E402
from System.Drawing import Color  # noqa: E402
from System.Globalization import CultureInfo  # noqa: E402
from System.Threading.Tasks import Task, TaskScheduler  # noqa: E402

clr.AddReference(interop_dll_path('Microsoft.Web.WebView2.Core.dll'))
clr.AddReference(interop_dll_path('Microsoft.Web.WebView2.WinForms.dll'))

from Microsoft.Web.WebView2.Core import (  # noqa: E402
    CoreWebView2Cookie,
    CoreWebView2ServerCertificateErrorAction,
    CoreWebView2WebResourceContext,
)
from Microsoft.Web.WebView2.WinForms import CoreWebView2CreationProperties, WebView2  # noqa: E402

for platform in ('win-arm64', 'win-x64', 'win-x86'):
    os.environ['Path'] += ';' + interop_dll_path(platform)


logger = logging.getLogger('pywebview')
renderer = 'edgechromium'


class EdgeChrome:
    def __init__(self, form: WinForms.Form, window: Window, cache_dir: str):
        self.pywebview_window = window
        self.webview = WebView2()
        props = CoreWebView2CreationProperties()

        runtime_path = webview_settings['WEBVIEW2_RUNTIME_PATH']
        if runtime_path:
            # Support relative paths
            if not os.path.isabs(runtime_path):
                runtime_path = os.path.join(get_app_root(), runtime_path)

            # Validate path exists
            if os.path.exists(runtime_path):
                props.BrowserExecutableFolder = runtime_path
                logger.debug(f'Using custom WebView2 runtime: {runtime_path}')
            else:
                logger.warning(
                    f'Custom WebView2 runtime path does not exist: {runtime_path}. Using system WebView2.'
                )

        props.UserDataFolder = cache_dir
        self.user_data_folder = props.UserDataFolder
        props.set_IsInPrivateModeEnabled(_state['private_mode'])
        props.AdditionalBrowserArguments = '--disable-features=ElasticOverscroll'

        if webview_settings['ALLOW_FILE_URLS']:
            props.AdditionalBrowserArguments += ' --allow-file-access-from-files'

        if webview_settings['REMOTE_DEBUGGING_PORT'] is not None:
            props.AdditionalBrowserArguments += (
                f' --remote-debugging-port={webview_settings["REMOTE_DEBUGGING_PORT"]}'
            )

        self.webview.CreationProperties = props

        self.form = form
        form.Controls.Add(self.webview)

        self.js_results = {}
        self.js_result_semaphore = Semaphore(0)
        self.webview.Dock = WinForms.DockStyle.Fill
        self.webview.BringToFront()
        self.webview.CoreWebView2InitializationCompleted += self.on_webview_ready
        self.webview.NavigationStarting += self.on_navigation_start
        self.webview.NavigationCompleted += self.on_navigation_completed
        self.webview.WebMessageReceived += self.on_script_notify
        self.syncContextTaskScheduler = TaskScheduler.FromCurrentSynchronizationContext()
        self.webview.DefaultBackgroundColor = Color.FromArgb(
            255,
            int(window.background_color.lstrip('#')[0:2], 16),
            int(window.background_color.lstrip('#')[2:4], 16),
            int(window.background_color.lstrip('#')[4:6], 16),
        )

        if window.transparent:
            self.webview.DefaultBackgroundColor = Color.Transparent

        self.url = None
        self.ishtml = False
        self.html = DEFAULT_HTML

        self.webview.EnsureCoreWebView2Async(None)

    def clear_user_data(self):
        if not _state['private_mode']:
            return

        try:
            process_id = Convert.ToInt32(self.webview.CoreWebView2.BrowserProcessId)
            process = Process.GetProcessById(process_id)
            self.webview.Dispose()
            process.WaitForExit(3000)

            shutil.rmtree(self.user_data_folder)
        except Exception as e:
            logger.warning(f'Failed to delete user data folder: {e}')

    def evaluate_js(self, script: str, parse_json: bool):
        def _callback(res):
            nonlocal result
            if parse_json and res is not None:
                try:
                    result = json.loads(res)
                except Exception:
                    result = res
            else:
                result = res
            semaphore.release()

        result = None
        semaphore = Semaphore(0)

        try:
            self.webview.Invoke(
                Func[Object](
                    lambda: self.webview.ExecuteScriptAsync(script).ContinueWith(
                        Action[Task[String]](lambda task: _callback(json.loads(task.Result))),
                        self.syncContextTaskScheduler,
                    )
                )
            )
            semaphore.acquire()
        except Exception:
            logger.exception('Error occurred in script')
            semaphore.release()

        return result

    def clear_cookies(self):
        self.webview.CoreWebView2.CookieManager.DeleteAllCookies()

    def get_cookies(self, cookies, semaphore):
        def _callback(task):
            for c in task.Result:
                _cookies.append(c)

            self.webview.Invoke(Func[Type](_parse_cookies))

        def _parse_cookies():
            # cookies must be accessed in the main thread, otherwise an exception is thrown
            # https://github.com/MicrosoftEdge/WebView2Feedback/issues/1976
            for c in _cookies:
                same_site = None if c.SameSite == 0 else str(c.SameSite).lower()
                try:
                    data = {
                        'name': c.Name,
                        'value': c.Value,
                        'path': c.Path,
                        'domain': c.Domain,
                        'expires': c.Expires.ToString('r', CultureInfo.GetCultureInfo('en-US')),
                        'secure': c.IsSecure,
                        'httponly': c.IsHttpOnly,
                        'samesite': same_site,
                    }

                    cookie = create_cookie(data)
                    cookies.append(cookie)
                except Exception as e:
                    logger.exception(e)

            semaphore.release()

        _cookies = []
        self.webview.CoreWebView2.CookieManager.GetCookiesAsync(self.url).ContinueWith(
            Action[Task[List[CoreWebView2Cookie]]](_callback), self.syncContextTaskScheduler
        )

    def get_current_url(self):
        return self.url

    def load_html(self, content, _):
        self.html = content
        self.ishtml = True

        if self.webview.CoreWebView2:
            self.webview.CoreWebView2.NavigateToString(self.html)
        else:
            self.webview.EnsureCoreWebView2Async(None)

    def load_url(self, url: str):
        self.ishtml = False
        self.webview.Source = Uri(url)

    def on_certificate_error(self, _, args):
        args.set_Action(CoreWebView2ServerCertificateErrorAction.AlwaysAllow)

    def on_script_notify(self, _, args):
        try:
            return_value = args.get_WebMessageAsJson()

            if return_value == '"FilesDropped"':
                if _dnd_state['num_listeners'] == 0:
                    return
                additionalObjects = args.get_AdditionalObjects()
                if additionalObjects is None:
                    return

                files = [
                    (os.path.basename(file.Path), file.Path)
                    for file in list(additionalObjects)
                    if 'CoreWebView2File' in str(type(file))
                ]
                _dnd_state['paths'] += files
                return

            func_name, func_param, value_id = json.loads(return_value)
            func_param = json.loads(func_param)
            if func_name == '_pywebviewAlert':
                WinForms.MessageBox.Show(str(func_param))
            elif func_name == 'console':
                print(func_param)
            else:
                js_bridge_call(self.pywebview_window, func_name, func_param, value_id)
        except Exception:
            logger.exception('Exception occurred during on_script_notify')

    def on_new_window_request(self, sender, args):
        args.set_Handled(True)

        if webview_settings['OPEN_EXTERNAL_LINKS_IN_BROWSER']:
            webbrowser.open(str(args.get_Uri()))
        else:
            self.load_url(str(args.get_Uri()))

    def on_source_changed(self, sender, args):
        self.url = sender.Source
        self.ishtml = False

    def on_webview_ready(self, sender, args):
        if not args.IsSuccess:
            logger.error(
                'WebView2 initialization failed with exception:\n '
                + str(args.InitializationException)
            )
            return

        self.webview.CoreWebView2.SourceChanged += self.on_source_changed
        sender.CoreWebView2.NewWindowRequested += self.on_new_window_request
        sender.CoreWebView2.AddWebResourceRequestedFilter('*', CoreWebView2WebResourceContext.All)
        sender.CoreWebView2.WebResourceResponseReceived += self.on_web_resource_response
        sender.CoreWebView2.WebResourceRequested += self.on_web_resource_request

        if _state['ssl'] or webview_settings['IGNORE_SSL_ERRORS']:
            sender.CoreWebView2.ServerCertificateErrorDetected += self.on_certificate_error

        sender.CoreWebView2.DownloadStarting += self.on_download_starting

        settings = sender.CoreWebView2.Settings
        settings.AreBrowserAcceleratorKeysEnabled = _state['debug']
        settings.AreDefaultContextMenusEnabled = _state['debug']
        settings.AreDefaultScriptDialogsEnabled = True
        settings.AreDevToolsEnabled = _state['debug']
        settings.IsBuiltInErrorPageEnabled = True
        settings.IsScriptEnabled = True
        settings.IsWebMessageEnabled = True
        settings.IsStatusBarEnabled = _state['debug']
        settings.IsSwipeNavigationEnabled = False
        settings.IsZoomControlEnabled = True

        if _state['user_agent']:
            settings.UserAgent = _state['user_agent']

        if _state['private_mode']:
            # cookies persist even if UserDataFolder is in memory. We have to delete cookies manually.
            sender.CoreWebView2.CookieManager.DeleteAllCookies()

        if self.pywebview_window.real_url:
            self.load_url(self.pywebview_window.real_url)
        elif self.pywebview_window.html:
            self.html = self.pywebview_window.html
            self.load_html(self.pywebview_window.html, '')
        else:
            self.load_html(DEFAULT_HTML, '')

        if _state['debug'] and webview_settings['OPEN_DEVTOOLS_IN_DEBUG']:
            sender.CoreWebView2.OpenDevToolsWindow()

    def on_download_starting(self, sender, args):
        if not webview_settings['ALLOW_DOWNLOADS']:
            args.Cancel = True
            return

        dialog = WinForms.SaveFileDialog()

        try:
            with winreg.OpenKey(
                winreg.HKEY_CURRENT_USER,
                r'Software\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders',
            ) as windows_key:
                dialog.InitialDirectory = winreg.QueryValueEx(
                    windows_key, '{374DE290-123F-4565-9164-39C4925E467B}'
                )[0]
        except Exception as e:
            logger.exception(e)

        dialog.Filter = (
            self.pywebview_window.localization['windows.fileFilter.allFiles'] + ' (*.*)|*.*'
        )
        dialog.RestoreDirectory = True
        dialog.FileName = os.path.basename(args.ResultFilePath)

        result = dialog.ShowDialog(self.form)
        if result == WinForms.DialogResult.OK:
            args.ResultFilePath = dialog.FileName
        else:
            args.Cancel = True

    def on_navigation_start(self, sender, args):
        if self.pywebview_window.transparent:
            self.form.Show()
            self.form.Activate()

    def on_web_resource_response(self, sender, args):
        headers = {}
        for header in args.Response.Headers.GetEnumerator():
            headers[header.Key] = header.Value

        response = Response(str(args.Request.Uri), args.Response.StatusCode, headers)
        self.pywebview_window.events.response_received.set(response)

    def on_web_resource_request(self, sender, args):
        original_headers = {}
        for header in args.Request.Headers.GetEnumerator():
            original_headers[header.Key] = header.Value

        request = Request(str(args.Request.Uri), args.Request.Method, original_headers)
        self.pywebview_window.events.request_sent.set(request)

        if request.headers == original_headers:
            return

        extra_headers = {
            k: v
            for k, v in request.headers.items()
            if k not in original_headers or original_headers[k] != v
        }
        missing_headers = {
            k: str(v) for k, v in original_headers.items() if k not in request.headers
        }

        for k, v in extra_headers.items():
            args.Request.Headers.SetHeader(k, v)

        for k in missing_headers:
            args.Request.Headers.RemoveHeader(k)

    def on_navigation_completed(self, sender, _):
        url = str(sender.Source)
        self.url = None if self.ishtml else url

        inject_pywebview(renderer, self.pywebview_window)



__all__ = ('KeyListener', 'FrameCallback')

from jnius import PythonJavaClass, java_method


class KeyListener(PythonJavaClass):
    """
    Represents a Python implementation of an Android `OnKeyListener`.

    This class provides an interface for handling key events in an Android environment
    using Python. It acts as a bridge between Python and the Android KeyListener
    interface. The primary purpose is to allow the execution of a Python function
    each time a key event is triggered.

    The class is structured to integrate with the Java ecosystem and is instantiated
    with a listener function defined in Python. This function enables the handling of
    key events by passing relevant details to it, empowering flexible event handling
    in a Pythonic manner.

    Attributes:
    listener (Callable): A Python callable function that will be triggered
        when a key event occurs. The function is expected to have three parameters:
        - v (android.view.View): The view that triggered the key event.
        - key_code (int): The key code of the key event.
        - event (android.view.KeyEvent): The Android KeyEvent object.
    """

    __javacontext__ = 'app'
    __javainterfaces__ = ['android/view/View$OnKeyListener']

    def __init__(self, listener):
        self.listener = listener

    @java_method('(Landroid/view/View;ILandroid/view/KeyEvent;)Z')
    def onKey(self, v, key_code, event):
        return self.listener(v, key_code, event)


class FrameCallback(PythonJavaClass):
    __javacontext__ = 'app'
    __javainterfaces__ = ['android/view/Choreographer$FrameCallback']

    def __init__(self, do_frame):
        self.do_frame = do_frame

    @java_method('(J)V')
    def doFrame(self, frame_time_nanos):
        self.do_frame(frame_time_nanos)
